import cv2
import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset
from torch.utils.data.dataset import Subset
from torchvision.datasets.sbd import SBDataset
from torchvision.datasets.voc import VOCSegmentation

import ignite.distributed as idist
from ignite.utils import convert_tensor


class TransformedDataset(Dataset):
    def __init__(self, ds, transform_fn):
        assert isinstance(ds, Dataset)
        assert callable(transform_fn)
        self.ds = ds
        self.transform_fn = transform_fn

    def __len__(self):
        return len(self.ds)

    def __getitem__(self, index):
        dp = self.ds[index]
        return self.transform_fn(**dp)


class VOCSegmentationOpencv(VOCSegmentation):
    target_names = [
        "background",
        "aeroplane",
        "bicycle",
        "bird",
        "boat",
        "bottle",
        "bus",
        "car",
        "cat",
        "chair",
        "cow",
        "diningtable",
        "dog",
        "horse",
        "motorbike",
        "person",
        "plant",
        "sheep",
        "sofa",
        "train",
        "tv/monitor",
    ]

    def __init__(self, *args, return_meta=False, **kwargs):
        super(VOCSegmentationOpencv, self).__init__(*args, **kwargs)
        self.return_meta = return_meta

    def __getitem__(self, index):
        img = cv2.imread(self.images[index])
        assert img is not None, f"Image at '{self.images[index]}' has a problem"
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        mask = np.asarray(Image.open(self.masks[index]))

        if self.return_meta:
            return {
                "image": img,
                "mask": mask,
                "meta": {"index": index, "image_path": self.images[index], "mask_path": self.masks[index]},
            }

        return {"image": img, "mask": mask}


class SBDatasetOpencv(SBDataset):
    def __init__(self, *args, return_meta=False, **kwargs):
        super(SBDatasetOpencv, self).__init__(*args, **kwargs)
        assert self.mode == "segmentation", "SBDatasetOpencv should be in segmentation mode only"
        self.return_meta = return_meta

    def _get_segmentation_target(self, filepath):
        mat = self._loadmat(filepath)
        return mat["GTcls"][0]["Segmentation"][0]

    def __getitem__(self, index):
        img = cv2.imread(self.images[index])
        assert img is not None, f"Image at '{self.images[index]}' has a problem"
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        mask = self._get_target(self.masks[index])

        if self.return_meta:
            return {
                "image": img,
                "mask": mask,
                "meta": {"index": index, "image_path": self.images[index], "mask_path": self.masks[index]},
            }

        return {"image": img, "mask": mask}


def get_train_dataset(root_path, return_meta=False):
    return VOCSegmentationOpencv(
        root=root_path, year="2012", image_set="train", download=False, return_meta=return_meta
    )


def get_val_dataset(root_path, return_meta=False):
    return VOCSegmentationOpencv(root=root_path, year="2012", image_set="val", download=False, return_meta=return_meta)


def get_train_noval_sbdataset(root_path, return_meta=False):
    return SBDatasetOpencv(root_path, image_set="train_noval", mode="segmentation", return_meta=return_meta)


def get_dataloader(dataset, sampler=None, shuffle=False, limit_num_samples=None, **kwargs):
    if limit_num_samples is not None:
        g = torch.Generator().manual_seed(limit_num_samples)
        indices = torch.randperm(len(dataset), generator=g)[:limit_num_samples]
        dataset = Subset(dataset, indices)

    return idist.auto_dataloader(dataset, sampler=sampler, shuffle=(sampler is None) and shuffle, **kwargs)


def get_train_val_loaders(
    root_path,
    train_transforms,
    val_transforms,
    batch_size=16,
    num_workers=8,
    train_sampler=None,
    val_batch_size=None,
    sbd_path=None,
    limit_train_num_samples=None,
    limit_val_num_samples=None,
):
    train_ds = get_train_dataset(root_path)
    val_ds = get_val_dataset(root_path)

    if sbd_path is not None:
        sbd_train_ds = get_train_noval_sbdataset(sbd_path)
        train_ds = train_ds + sbd_train_ds

    if len(val_ds) < len(train_ds):
        g = torch.Generator().manual_seed(len(train_ds))
        train_eval_indices = torch.randperm(len(train_ds), generator=g)[: len(val_ds)]
        train_eval_ds = Subset(train_ds, train_eval_indices)
    else:
        train_eval_ds = train_ds

    train_ds = TransformedDataset(train_ds, transform_fn=train_transforms)
    val_ds = TransformedDataset(val_ds, transform_fn=val_transforms)
    train_eval_ds = TransformedDataset(train_eval_ds, transform_fn=val_transforms)

    val_batch_size = batch_size * 4 if val_batch_size is None else val_batch_size

    train_loader = get_dataloader(
        train_ds,
        shuffle=True,
        sampler=train_sampler,
        batch_size=batch_size,
        num_workers=num_workers,
        drop_last=True,
        limit_num_samples=limit_train_num_samples,
    )

    val_loader = get_dataloader(
        val_ds,
        shuffle=False,
        batch_size=val_batch_size,
        num_workers=num_workers,
        drop_last=False,
        limit_num_samples=limit_val_num_samples,
    )

    train_eval_loader = get_dataloader(
        train_eval_ds,
        shuffle=False,
        batch_size=val_batch_size,
        num_workers=num_workers,
        drop_last=False,
        limit_num_samples=limit_val_num_samples,
    )

    return train_loader, val_loader, train_eval_loader


def get_inference_dataloader(
    root_path, mode, transforms, batch_size=16, num_workers=8, pin_memory=True, limit_num_samples=None
):
    assert mode in ("train", "test"), "Mode should be 'train' or 'test'"

    get_dataset_fn = get_train_dataset if mode == "train" else get_val_dataset

    dataset = get_dataset_fn(root_path, return_meta=True)
    dataset = TransformedDataset(dataset, transform_fn=transforms)
    return get_dataloader(
        dataset,
        limit_num_samples=limit_num_samples,
        shuffle=False,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=pin_memory,
        drop_last=False,
    )


def ignore_mask_boundaries(**kwargs):
    assert "mask" in kwargs, "Input should contain 'mask'"
    mask = kwargs["mask"]
    mask[mask == 255] = 0
    kwargs["mask"] = mask
    return kwargs


def denormalize(t, mean, std, max_pixel_value=255):
    assert isinstance(t, torch.Tensor), f"{type(t)}"
    assert t.ndim == 3
    d = t.device
    mean = torch.tensor(mean, device=d).unsqueeze(-1).unsqueeze(-1)
    std = torch.tensor(std, device=d).unsqueeze(-1).unsqueeze(-1)
    tensor = std * t + mean
    tensor *= max_pixel_value
    return tensor


def prepare_image_mask(batch, device, non_blocking):
    x, y = batch["image"], batch["mask"]
    x = convert_tensor(x, device, non_blocking=non_blocking)
    y = convert_tensor(y, device, non_blocking=non_blocking).long()
    return x, y



import torch


class TransformerDataset(torch.utils.data.Dataset):
    def __init__(self, texts, labels, tokenizer, max_length):
        self.texts = texts
        self.labels = labels
        self.tokenizer = tokenizer
        self.max_length = max_length

    def __getitem__(self, idx):
        text = str(self.texts[idx])
        text = " ".join(text.split())
        inputs = self.tokenizer(
            text,
            None,
            add_special_tokens=True,
            max_length=self.max_length,
            truncation=True,
            padding="max_length",
            return_tensors="pt",
        )
        inputs = {k: v.type(torch.long).squeeze(0) for k, v in inputs.items()}

        labels_pt = torch.tensor(self.labels[idx], dtype=torch.float)
        return inputs, labels_pt

    def __len__(self):
        return len(self.labels)



import random

from torch.utils.data import DataLoader, Subset
from torchvision.datasets.cifar import CIFAR100
from torchvision.transforms import Compose, Normalize, Pad, RandomCrop, RandomErasing, RandomHorizontalFlip, ToTensor


def get_train_eval_loaders(path, batch_size=256):
    """Setup the dataflow:
        - load CIFAR100 train and test datasets
        - setup train/test image transforms
            - horizontally flipped randomly and augmented using cutout.
            - each mini-batch contained 256 examples
        - setup train/test data loaders

    Returns:
        train_loader, test_loader, eval_train_loader
    """
    train_transform = Compose(
        [
            Pad(4),
            RandomCrop(32),
            RandomHorizontalFlip(),
            ToTensor(),
            Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            RandomErasing(),
        ]
    )

    test_transform = Compose([ToTensor(), Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])

    train_dataset = CIFAR100(root=path, train=True, transform=train_transform, download=True)
    test_dataset = CIFAR100(root=path, train=False, transform=test_transform, download=False)

    train_eval_indices = [random.randint(0, len(train_dataset) - 1) for i in range(len(test_dataset))]
    train_eval_dataset = Subset(train_dataset, train_eval_indices)

    train_loader = DataLoader(
        train_dataset, batch_size=batch_size, num_workers=12, shuffle=True, drop_last=True, pin_memory=True
    )

    test_loader = DataLoader(
        test_dataset, batch_size=batch_size, num_workers=12, shuffle=False, drop_last=False, pin_memory=True
    )

    eval_train_loader = DataLoader(
        train_eval_dataset, batch_size=batch_size, num_workers=12, shuffle=False, drop_last=False, pin_memory=True
    )

    return train_loader, test_loader, eval_train_loader



import os

import numpy as np
import pytest
import torch

import ignite.distributed as idist
from ignite.exceptions import NotComputableError
from ignite.metrics import RootMeanSquaredError


def test_zero_sample():
    rmse = RootMeanSquaredError()
    with pytest.raises(
        NotComputableError, match=r"MeanSquaredError must have at least one example before it can be computed"
    ):
        rmse.compute()


@pytest.fixture(params=[0, 1, 2, 3])
def test_data(request):
    return [
        (torch.empty(10).uniform_(0, 10), torch.empty(10).uniform_(0, 10), 1),
        (torch.empty(10, 1).uniform_(-10, 10), torch.empty(10, 1).uniform_(-10, 10), 1),
        # updated batches
        (torch.empty(50).uniform_(0, 10), torch.empty(50).uniform_(0, 10), 16),
        (torch.empty(50, 1).uniform_(-10, 10), torch.empty(50, 1).uniform_(-10, 10), 16),
    ][request.param]


@pytest.mark.parametrize("n_times", range(3))
def test_compute(n_times, test_data, available_device):
    rmse = RootMeanSquaredError(device=available_device)
    assert rmse._device == torch.device(available_device)

    y_pred, y, batch_size = test_data
    rmse.reset()
    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            rmse.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        rmse.update((y_pred, y))

    np_y = y.numpy().ravel()
    np_y_pred = y_pred.numpy().ravel()

    np_res = np.sqrt(np.power((np_y - np_y_pred), 2.0).sum() / np_y.shape[0])
    res = rmse.compute()

    assert isinstance(res, float)
    assert pytest.approx(res) == np_res


def _test_distrib_integration(device, tol=1e-6):
    from ignite.engine import Engine

    rank = idist.get_rank()

    def _test(metric_device):
        n_iters = 2
        batch_size = 3

        torch.manual_seed(12 + rank)

        y_true = torch.arange(0, n_iters * batch_size, dtype=torch.float).to(device)
        y_preds = (rank + 1) * torch.ones(n_iters * batch_size, dtype=torch.float).to(device)

        def update(engine, i):
            return y_preds[i * batch_size : (i + 1) * batch_size], y_true[i * batch_size : (i + 1) * batch_size]

        engine = Engine(update)

        m = RootMeanSquaredError(device=metric_device)
        m.attach(engine, "rmse")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=1)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "rmse" in engine.state.metrics
        res = engine.state.metrics["rmse"]

        true_res = np.sqrt(np.mean(np.square((y_true - y_preds).cpu().numpy())))

        assert pytest.approx(res, rel=tol) == true_res

    _test("cpu")
    if device.type != "xla":
        _test(idist.device())


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_integration(device, tol=1e-4)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_integration(device, tol=1e-4)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



from typing import Callable, Tuple, Union

import torch

from ignite.exceptions import NotComputableError
from ignite.metrics.metric import reinit__is_reduced, sync_all_reduce

from ignite.metrics.regression._base import _BaseRegression


class PearsonCorrelation(_BaseRegression):
    r"""Calculates the
    `Pearson correlation coefficient <https://en.wikipedia.org/wiki/Pearson_correlation_coefficient>`_.

    .. math::
        r = \frac{\sum_{j=1}^n (P_j-\bar{P})(A_j-\bar{A})}
        {\max (\sqrt{\sum_{j=1}^n (P_j-\bar{P})^2 \sum_{j=1}^n (A_j-\bar{A})^2}, \epsilon)},
        \quad \bar{P}=\frac{1}{n}\sum_{j=1}^n P_j, \quad \bar{A}=\frac{1}{n}\sum_{j=1}^n A_j

    where :math:`A_j` is the ground truth and :math:`P_j` is the predicted value.

    - ``update`` must receive output of the form ``(y_pred, y)`` or ``{'y_pred': y_pred, 'y': y}``.
    - `y` and `y_pred` must be of same shape `(N, )` or `(N, 1)`.

    Parameters are inherited from ``Metric.__init__``.

    Args:
        eps: a small value to avoid division by zero. Default: 1e-8
        output_transform: a callable that is used to transform the
            :class:`~ignite.engine.engine.Engine`'s ``process_function``'s output into the
            form expected by the metric. This can be useful if, for example, you have a multi-output model and
            you want to compute the metric with respect to one of the outputs.
            By default, metrics require the output as ``(y_pred, y)`` or ``{'y_pred': y_pred, 'y': y}``.
        device: specifies which device updates are accumulated on. Setting the
            metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
            non-blocking. By default, CPU.

    Examples:
        To use with ``Engine`` and ``process_function``, simply attach the metric instance to the engine.
        The output of the engine's ``process_function`` needs to be in format of
        ``(y_pred, y)`` or ``{'y_pred': y_pred, 'y': y, ...}``.

        .. include:: defaults.rst
            :start-after: :orphan:

        .. testcode::

            metric = PearsonCorrelation()
            metric.attach(default_evaluator, 'corr')
            y_true = torch.tensor([0., 1., 2., 3., 4., 5.])
            y_pred = torch.tensor([0.5, 1.3, 1.9, 2.8, 4.1, 6.0])
            state = default_evaluator.run([[y_pred, y_true]])
            print(state.metrics['corr'])

        .. testoutput::

            0.9768688678741455
    """

    def __init__(
        self,
        eps: float = 1e-8,
        output_transform: Callable = lambda x: x,
        device: Union[str, torch.device] = torch.device("cpu"),
    ):
        super().__init__(output_transform, device)

        self.eps = eps

    _state_dict_all_req_keys = (
        "_sum_of_y_preds",
        "_sum_of_ys",
        "_sum_of_y_pred_squares",
        "_sum_of_y_squares",
        "_sum_of_products",
        "_num_examples",
    )

    @reinit__is_reduced
    def reset(self) -> None:
        self._sum_of_y_preds = torch.tensor(0.0, device=self._device)
        self._sum_of_ys = torch.tensor(0.0, device=self._device)
        self._sum_of_y_pred_squares = torch.tensor(0.0, device=self._device)
        self._sum_of_y_squares = torch.tensor(0.0, device=self._device)
        self._sum_of_products = torch.tensor(0.0, device=self._device)
        self._num_examples = 0

    def _update(self, output: Tuple[torch.Tensor, torch.Tensor]) -> None:
        y_pred, y = output[0].detach(), output[1].detach()
        self._sum_of_y_preds += y_pred.sum().to(self._device)
        self._sum_of_ys += y.sum().to(self._device)
        self._sum_of_y_pred_squares += y_pred.square().sum().to(self._device)
        self._sum_of_y_squares += y.square().sum().to(self._device)
        self._sum_of_products += (y_pred * y).sum().to(self._device)
        self._num_examples += y.shape[0]

    @sync_all_reduce(
        "_sum_of_y_preds",
        "_sum_of_ys",
        "_sum_of_y_pred_squares",
        "_sum_of_y_squares",
        "_sum_of_products",
        "_num_examples",
    )
    def compute(self) -> float:
        n = self._num_examples
        if n == 0:
            raise NotComputableError("PearsonCorrelation must have at least one example before it can be computed.")

        # cov = E[xy] - E[x]*E[y]
        cov = self._sum_of_products / n - self._sum_of_y_preds * self._sum_of_ys / (n * n)

        # var = E[x^2] - E[x]^2
        y_pred_mean = self._sum_of_y_preds / n
        y_pred_var = self._sum_of_y_pred_squares / n - y_pred_mean * y_pred_mean
        y_pred_var = torch.clamp(y_pred_var, min=0.0)

        y_mean = self._sum_of_ys / n
        y_var = self._sum_of_y_squares / n - y_mean * y_mean
        y_var = torch.clamp(y_var, min=0.0)

        r = cov / torch.clamp(torch.sqrt(y_pred_var * y_var), min=self.eps)
        return float(r.item())



from typing import Optional

import numpy as np

import pytest
import torch

import ignite.distributed as idist

from ignite.metrics.regression._base import _BaseRegression, _torch_median


def test_base_regression_shapes():
    class L1(_BaseRegression):
        def reset(self):
            self._sum_of_errors = 0.0

        def _update(self, output):
            y_pred, y = output
            errors = torch.abs(y.view_as(y_pred) - y_pred)
            self._sum_of_errors += torch.sum(errors).item()

        def compute(self):
            return self._sum_of_errors

    m = L1()

    with pytest.raises(ValueError, match=r"Input y_pred should have shape \(N,\) or \(N, 1\)"):
        y = torch.rand([1, 1, 1])
        m.update((y, y))

    with pytest.raises(ValueError, match=r"Input y should have shape \(N,\) or \(N, 1\)"):
        y = torch.rand([1, 1, 1])
        m.update((torch.rand(1, 1), y))

    with pytest.raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(2), torch.rand(2, 1)))

    with pytest.raises(TypeError, match=r"Input y_pred dtype should be float"):
        y = torch.tensor([1, 1])
        m.update((y, y))

    with pytest.raises(TypeError, match=r"Input y dtype should be float"):
        y = torch.tensor([1, 1])
        m.update((y.float(), y))


@pytest.mark.parametrize("size", [100, 101, (30, 3), (31, 3)])
def test_torch_median_numpy(size, device: Optional[str] = None):
    data = torch.rand(size).to(device)
    assert _torch_median(data) == np.median(data.cpu().numpy())


@pytest.mark.tpu
@pytest.mark.parametrize("size", [100, 101, (30, 3), (31, 3)])
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_on_even_size_xla(size):
    device = "xla"
    test_torch_median_numpy(size, device=device)


@pytest.mark.parametrize("size", [100, 101, (30, 3), (31, 3)])
@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_on_even_size_gpu(size):
    test_torch_median_numpy(size, device="cuda")


@pytest.mark.parametrize("size", [100, 101, (30, 3), (31, 3)])
def test_create_even_size_cpu(size):
    test_torch_median_numpy(size, device="cpu")



from ignite.metrics.regression import (
    canberra_metric,
    fractional_absolute_error,
    fractional_bias,
    geometric_mean_absolute_error,
    geometric_mean_relative_absolute_error,
    manhattan_distance,
    maximum_absolute_error,
    mean_absolute_relative_error,
    mean_error,
    mean_normalized_bias,
    median_absolute_error,
    median_absolute_percentage_error,
    median_relative_absolute_error,
    r2_score,
    wave_hedges_distance,
)
from ignite.metrics.regression.canberra_metric import CanberraMetric
from ignite.metrics.regression.fractional_absolute_error import FractionalAbsoluteError
from ignite.metrics.regression.fractional_bias import FractionalBias
from ignite.metrics.regression.geometric_mean_absolute_error import GeometricMeanAbsoluteError
from ignite.metrics.regression.geometric_mean_relative_absolute_error import GeometricMeanRelativeAbsoluteError
from ignite.metrics.regression.manhattan_distance import ManhattanDistance
from ignite.metrics.regression.maximum_absolute_error import MaximumAbsoluteError
from ignite.metrics.regression.mean_absolute_relative_error import MeanAbsoluteRelativeError
from ignite.metrics.regression.mean_error import MeanError
from ignite.metrics.regression.mean_normalized_bias import MeanNormalizedBias
from ignite.metrics.regression.median_absolute_error import MedianAbsoluteError
from ignite.metrics.regression.median_absolute_percentage_error import MedianAbsolutePercentageError
from ignite.metrics.regression.median_relative_absolute_error import MedianRelativeAbsoluteError
from ignite.metrics.regression.r2_score import R2Score
from ignite.metrics.regression.wave_hedges_distance import WaveHedgesDistance


__all__ = [
    "CanberraMetric",
    "FractionalAbsoluteError",
    "FractionalBias",
    "GeometricMeanAbsoluteError",
    "GeometricMeanRelativeAbsoluteError",
    "ManhattanDistance",
    "MaximumAbsoluteError",
    "MeanAbsoluteRelativeError",
    "MeanError",
    "MeanNormalizedBias",
    "MedianAbsoluteError",
    "MedianAbsolutePercentageError",
    "MedianRelativeAbsoluteError",
    "R2Score",
    "WaveHedgesDistance",
    "canberra_metric",
    "fractional_absolute_error",
    "fractional_bias",
    "geometric_mean_absolute_error",
    "geometric_mean_relative_absolute_error",
    "manhattan_distance",
    "maximum_absolute_error",
    "mean_absolute_relative_error",
    "mean_error",
    "mean_normalized_bias",
    "median_absolute_error",
    "median_absolute_percentage_error",
    "median_relative_absolute_error",
    "r2_score",
    "wave_hedges_distance",
]



import argparse

import torch

import ignite.distributed as idist


def training(local_rank, config, **kwargs):
    import time

    time.sleep(idist.get_rank() * 0.1)

    print(idist.get_rank(), ": run with config:", config, "- kwargs:", kwargs, f"- backend={idist.backend()}")

    t = torch.tensor([idist.get_rank()], device=idist.device())
    t = idist.all_reduce(t)
    t = t.item()
    ws = idist.get_world_size()
    assert t == ws * (ws - 1) / 2, f"{t} vs {ws}"
    assert local_rank == idist.get_local_rank()

    # Test init method:
    if idist.model_name() == "native-dist":
        from ignite.distributed.utils import _model

        true_init_method = config.get("true_init_method", None)
        assert true_init_method is not None, true_init_method
        assert _model._init_method == true_init_method


if __name__ == "__main__":
    """
    Usage:

        - No distributed configuration:
        ```
        python tests/ignite/distributed/check_idist_parallel.py
        ```

        - Launch 4 procs using gloo backend with `torchrun`:

        ```
        torchrun --nproc_per_node=4 tests/ignite/distributed/check_idist_parallel.py --backend=gloo
        ```

        - Launch 2 procs in 2 nodes using gloo backend with `torchrun` or `torch.distributed.launch`:

        ```
        bash -c "torchrun --nnodes=2 --node_rank=0 \
            --master_addr=localhost --master_port=3344 --nproc_per_node=2 \
            tests/ignite/distributed/check_idist_parallel.py --backend=gloo &" \
        && bash -c "torchrun --nnodes=2 --node_rank=1 \
            --master_addr=localhost --master_port=3344 --nproc_per_node=2 \
            tests/ignite/distributed/check_idist_parallel.py --backend=gloo &"
        ```

        - Spawn 4 procs in single node using gloo backend:
        ```
        python tests/ignite/distributed/check_idist_parallel.py --backend=gloo --nproc_per_node=4
        ```

        - Spawn 2 procs in 2 nodes using gloo backend:
        ```
        bash -c "python tests/ignite/distributed/check_idist_parallel.py --backend=gloo \
            --nproc_per_node=2 --nnodes=2 --node_rank=0 --master_addr=localhost --master_port=3344 &" \
        && bash -c "python tests/ignite/distributed/check_idist_parallel.py --backend=gloo \
            --nproc_per_node=2 --nnodes=2 --node_rank=1 --master_addr=localhost --master_port=3344 &"
        ```

        - Spawn 8 procs in single node using xla-tpu backend:
        ```
        python tests/ignite/distributed/check_idist_parallel.py --backend=xla-tpu --nproc_per_node=8
        ```


    """

    parser = argparse.ArgumentParser("Check idist.Parallel")
    parser.add_argument("--backend", type=str, default=None)
    parser.add_argument("--nproc_per_node", type=int, default=None)
    parser.add_argument("--nnodes", type=int, default=None)
    parser.add_argument("--node_rank", type=int, default=None)
    parser.add_argument("--master_addr", type=str, default=None)
    parser.add_argument("--master_port", type=str, default=None)
    parser.add_argument("--init_method", type=str, default=None)

    args = parser.parse_args()

    config = {
        "model": "resnet18",
        "lr": 0.01,
    }
    if args.backend in ["gloo", "nccl"]:
        config["true_init_method"] = args.init_method if args.init_method is not None else "env://"

    dist_config = dict(
        nproc_per_node=args.nproc_per_node,
        nnodes=args.nnodes,
        node_rank=args.node_rank,
        master_addr=args.master_addr,
        master_port=args.master_port,
    )
    if args.init_method is not None:
        dist_config["init_method"] = args.init_method

    with idist.Parallel(backend=args.backend, **dist_config) as parallel:
        parallel.run(training, config, a=1, b=2)



import warnings

import pytest
import torch
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.metrics import recall_score

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics import Recall

torch.manual_seed(12)


def test_no_update():
    recall = Recall()
    assert recall._updated is False
    with pytest.raises(NotComputableError, match=r"Recall must have at least one example before it can be computed"):
        recall.compute()
    assert recall._updated is False

    recall = Recall(is_multilabel=True)
    assert recall._updated is False
    with pytest.raises(NotComputableError, match=r"Recall must have at least one example before it can be computed"):
        recall.compute()
    assert recall._updated is False


def test_average_parameter():
    re = Recall(average="samples")
    with pytest.raises(
        ValueError, match=r"Argument average='samples' is incompatible with binary and multiclass input data."
    ):
        re.update((torch.randint(0, 2, size=(10,)).long(), torch.randint(0, 2, size=(10,)).long()))
    assert re._updated is False

    re = Recall(average="samples")
    with pytest.raises(
        ValueError, match=r"Argument average='samples' is incompatible with binary and multiclass input data."
    ):
        re.update((torch.rand(10, 3), torch.randint(0, 3, size=(10,)).long()))
    assert re._updated is False

    re = Recall(average=True)
    assert re._average == "macro"


def test_binary_wrong_inputs():
    re = Recall()

    assert re._updated is False
    with pytest.raises(ValueError, match=r"For binary cases, y must be comprised of 0's and 1's"):
        # y has not only 0 or 1 values
        re.update((torch.randint(0, 2, size=(10,)), torch.arange(0, 10).long()))
    assert re._updated is False

    with pytest.raises(ValueError, match=r"For binary cases, y_pred must be comprised of 0's and 1's"):
        # y_pred values are not thresholded to 0, 1 values
        re.update((torch.rand(10, 1), torch.randint(0, 2, size=(10,)).long()))
    assert re._updated is False

    with pytest.raises(ValueError, match=r"y must have shape of"):
        # incompatible shapes
        re.update((torch.randint(0, 2, size=(10,)), torch.randint(0, 2, size=(10, 5)).long()))
    assert re._updated is False

    with pytest.raises(ValueError, match=r"y must have shape of"):
        # incompatible shapes
        re.update((torch.randint(0, 2, size=(10, 5, 6)), torch.randint(0, 2, size=(10,)).long()))
    assert re._updated is False

    with pytest.raises(ValueError, match=r"y must have shape of"):
        # incompatible shapes
        re.update((torch.randint(0, 2, size=(10,)), torch.randint(0, 2, size=(10, 5, 6)).long()))
    assert re._updated is False

    with pytest.warns(
        RuntimeWarning,
        match="`y` and `y_pred` should be of dtype long when entry type is binary and average!=False",
    ):
        re = Recall(average=None)
        re.update((torch.randint(0, 2, size=(10,)).float(), torch.randint(0, 2, size=(10,))))

    with pytest.warns(
        RuntimeWarning,
        match="`y` and `y_pred` should be of dtype long when entry type is binary and average!=False",
    ):
        re = Recall(average=None)
        re.update((torch.randint(0, 2, size=(10,)), torch.randint(0, 2, size=(10,)).float()))


def ignite_average_to_scikit_average(average, data_type: str):
    if average in [None, "micro", "samples", "weighted", "macro"]:
        return average
    if average is False:
        if data_type == "binary":
            return "binary"
        else:
            return None
    elif average is True:
        return "macro"
    else:
        raise ValueError(f"Wrong average parameter `{average}`")


@pytest.mark.parametrize("n_times", range(3))
@pytest.mark.parametrize("average", [None, False, "macro", "micro", "weighted"])
def test_binary_input(n_times, available_device, average, test_data_binary):
    re = Recall(average=average, device=available_device)
    assert re._device == torch.device(available_device)
    assert re._updated is False

    y_pred, y, batch_size = test_data_binary

    re.reset()
    assert re._updated is False

    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            re.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        re.update((y_pred, y))

    np_y = y.numpy().ravel()
    np_y_pred = y_pred.numpy().ravel()

    assert re._type == "binary"
    assert re._updated is True
    assert isinstance(re.compute(), torch.Tensor if not average else float)
    re_compute = re.compute().cpu().numpy() if not average else re.compute()
    sk_average_parameter = ignite_average_to_scikit_average(average, "binary")
    assert recall_score(np_y, np_y_pred, average=sk_average_parameter, labels=[0, 1]) == pytest.approx(re_compute)


def test_multiclass_wrong_inputs():
    re = Recall()
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes
        re.update((torch.rand(10, 5, 4), torch.randint(0, 2, size=(10,)).long()))
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes
        re.update((torch.rand(10, 5, 6), torch.randint(0, 5, size=(10, 5)).long()))
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes
        re.update((torch.rand(10), torch.randint(0, 5, size=(10, 5, 6)).long()))
    assert re._updated is False

    re = Recall(average=True)
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes between two updates
        re.update((torch.rand(10, 5), torch.randint(0, 5, size=(10,)).long()))
        re.update((torch.rand(10, 6), torch.randint(0, 5, size=(10,)).long()))
    assert re._updated is True

    with pytest.raises(ValueError):
        # incompatible shapes between two updates
        re.update((torch.rand(10, 5, 12, 14), torch.randint(0, 5, size=(10, 12, 14)).long()))
        re.update((torch.rand(10, 6, 12, 14), torch.randint(0, 5, size=(10, 12, 14)).long()))
    assert re._updated is True

    re = Recall(average=False)
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes between two updates
        re.update((torch.rand(10, 5), torch.randint(0, 5, size=(10,)).long()))
        re.update((torch.rand(10, 6), torch.randint(0, 5, size=(10,)).long()))
    assert re._updated is True

    with pytest.raises(ValueError):
        # incompatible shapes between two updates
        re.update((torch.rand(10, 5, 12, 14), torch.randint(0, 5, size=(10, 12, 14)).long()))
        re.update((torch.rand(10, 6, 12, 14), torch.randint(0, 5, size=(10, 12, 14)).long()))
    assert re._updated is True

    with pytest.warns(
        RuntimeWarning,
        match="`y` should be of dtype long when entry type is multiclass",
    ):
        re = Recall()
        re.update((torch.rand(10, 5), torch.randint(0, 5, size=(10,)).float()))


@pytest.mark.parametrize("n_times", range(3))
@pytest.mark.parametrize("average", [None, False, "macro", "micro", "weighted"])
def test_multiclass_input(n_times, available_device, average, test_data_multiclass):
    re = Recall(average=average, device=available_device)
    assert re._device == torch.device(available_device)
    assert re._updated is False

    y_pred, y, batch_size = test_data_multiclass
    re.reset()
    assert re._updated is False

    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            re.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        re.update((y_pred, y))

    num_classes = y_pred.shape[1]
    np_y_pred = y_pred.argmax(dim=1).numpy().ravel()
    np_y = y.numpy().ravel()

    assert re._type == "multiclass"
    assert re._updated is True
    assert isinstance(re.compute(), torch.Tensor if not average else float)
    re_compute = re.compute().cpu().numpy() if not average else re.compute()
    sk_average_parameter = ignite_average_to_scikit_average(average, "multiclass")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UndefinedMetricWarning)
        sk_compute = recall_score(np_y, np_y_pred, labels=range(0, num_classes), average=sk_average_parameter)
        assert sk_compute == pytest.approx(re_compute)


def test_multilabel_wrong_inputs():
    re = Recall(is_multilabel=True)
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes
        re.update((torch.randint(0, 2, size=(10,)), torch.randint(0, 2, size=(10,)).long()))
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible y_pred
        re.update((torch.rand(10, 5), torch.randint(0, 2, size=(10, 5)).long()))
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible y
        re.update((torch.randint(0, 5, size=(10, 5, 6)), torch.rand(10)))
    assert re._updated is False

    with pytest.raises(ValueError):
        # incompatible shapes between two updates
        re.update((torch.randint(0, 2, size=(20, 5)), torch.randint(0, 2, size=(20, 5)).long()))
        re.update((torch.randint(0, 2, size=(20, 6)), torch.randint(0, 2, size=(20, 6)).long()))
    assert re._updated is True


def to_numpy_multilabel(y):
    # reshapes input array to (N x ..., C)
    y = y.transpose(1, 0).cpu().numpy()
    num_classes = y.shape[0]
    y = y.reshape((num_classes, -1)).transpose(1, 0)
    return y


@pytest.mark.parametrize("n_times", range(3))
@pytest.mark.parametrize("average", [None, False, "macro", "micro", "weighted", "samples"])
def test_multilabel_input(n_times, available_device, average, test_data_multilabel):

    re = Recall(average=average, is_multilabel=True, device=available_device)
    assert re._device == torch.device(available_device)
    assert re._updated is False

    y_pred, y, batch_size = test_data_multilabel
    re.reset()
    assert re._updated is False

    if batch_size > 1:
        n_iters = y.shape[0] // batch_size + 1
        for i in range(n_iters):
            idx = i * batch_size
            re.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))
    else:
        re.update((y_pred, y))

    np_y_pred = to_numpy_multilabel(y_pred)
    np_y = to_numpy_multilabel(y)

    assert re._type == "multilabel"
    assert re._updated is True
    re_compute = re.compute().cpu().numpy() if not average else re.compute()
    sk_average_parameter = ignite_average_to_scikit_average(average, "multilabel")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UndefinedMetricWarning)
        assert recall_score(np_y, np_y_pred, average=sk_average_parameter) == pytest.approx(re_compute)


@pytest.mark.parametrize("average", [None, False, "macro", "micro", "weighted"])
def test_incorrect_type(average):
    # Tests changing of type during training

    re = Recall(average=average)
    assert re._updated is False

    y_pred = torch.softmax(torch.rand(4, 4), dim=1)
    y = torch.ones(4).long()
    re.update((y_pred, y))
    assert re._updated is True

    y_pred = torch.zeros(4)
    y = torch.ones(4).long()

    with pytest.raises(RuntimeError):
        re.update((y_pred, y))

    assert re._updated is True


@pytest.mark.parametrize("average", [None, False, "macro", "micro", "weighted"])
def test_incorrect_y_classes(average):
    re = Recall(average=average)

    assert re._updated is False

    y_pred = torch.randint(0, 2, size=(10, 4)).float()
    y = torch.randint(4, 5, size=(10,)).long()

    with pytest.raises(ValueError):
        re.update((y_pred, y))

    assert re._updated is False


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    @pytest.mark.parametrize("average", [False, "macro", "weighted", "micro"])
    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration_multiclass(self, average, n_epochs):
        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        n_iters = 60
        batch_size = 16
        n_classes = 7

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(idist.device())

        for metric_device in metric_devices:
            y_true = torch.randint(0, n_classes, size=(n_iters * batch_size,)).to(device)
            y_preds = torch.rand(n_iters * batch_size, n_classes).to(device)

            def update(engine, i):
                return (
                    y_preds[i * batch_size : (i + 1) * batch_size, :],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )

            engine = Engine(update)

            re = Recall(average=average, device=metric_device)
            re.attach(engine, "re")
            assert re._updated is False

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "re" in engine.state.metrics
            assert re._updated is True
            res = engine.state.metrics["re"]
            if isinstance(res, torch.Tensor):
                # Fixes https://github.com/pytorch/ignite/issues/1635#issuecomment-863026919
                assert res.device.type == "cpu"
                res = res.cpu().numpy()

            sk_average_parameter = ignite_average_to_scikit_average(average, "multiclass")
            true_res = recall_score(
                y_true.cpu().numpy(), torch.argmax(y_preds, dim=1).cpu().numpy(), average=sk_average_parameter
            )

            assert pytest.approx(res) == true_res

    @pytest.mark.parametrize("average", [False, "macro", "weighted", "micro", "samples"])
    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration_multilabel(self, average, n_epochs):
        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        n_iters = 60
        batch_size = 16
        n_classes = 7

        metric_devices = ["cpu"]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(idist.device())

        for metric_device in metric_devices:
            y_true = torch.randint(0, 2, size=(n_iters * batch_size, n_classes, 6, 8)).to(device)
            y_preds = torch.randint(0, 2, size=(n_iters * batch_size, n_classes, 6, 8)).to(device)

            def update(engine, i):
                return (
                    y_preds[i * batch_size : (i + 1) * batch_size, ...],
                    y_true[i * batch_size : (i + 1) * batch_size, ...],
                )

            engine = Engine(update)

            re = Recall(average=average, is_multilabel=True, device=metric_device)
            re.attach(engine, "re")
            assert re._updated is False

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "re" in engine.state.metrics
            assert re._updated is True
            res = engine.state.metrics["re"]
            res2 = re.compute()
            if isinstance(res, torch.Tensor):
                res = res.cpu().numpy()
                res2 = res2.cpu().numpy()
                assert (res == res2).all()
            else:
                assert res == res2

            np_y_preds = to_numpy_multilabel(y_preds)
            np_y_true = to_numpy_multilabel(y_true)
            assert re._type == "multilabel"
            sk_average_parameter = ignite_average_to_scikit_average(average, "multilabel")
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UndefinedMetricWarning)
                assert recall_score(np_y_true, np_y_preds, average=sk_average_parameter) == pytest.approx(res)

    @pytest.mark.parametrize("average", [False, "macro", "weighted", "micro"])
    def test_accumulator_device(self, average):
        # Binary accuracy on input of shape (N, 1) or (N, )

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(idist.device())
        for metric_device in metric_devices:
            re = Recall(average=average, device=metric_device)
            assert re._device == metric_device
            assert re._updated is False
            # Since the shape of the accumulated amount isn't known before the first update
            # call, the internal variables aren't tensors on the right device yet.

            y_reed = torch.randint(0, 2, size=(10,))
            y = torch.randint(0, 2, size=(10,)).long()
            re.update((y_reed, y))

            assert re._updated is True

            assert (
                re._numerator.device == metric_device
            ), f"{type(re._numerator.device)}:{re._numerator.device} vs {type(metric_device)}:{metric_device}"

            if average != "samples":
                # For average='samples', `_denominator` is of type `int` so it has not `device` member.
                assert (
                    re._denominator.device == metric_device
                ), f"{type(re._denominator.device)}:{re._denominator.device} vs {type(metric_device)}:{metric_device}"

            if average == "weighted":
                assert re._weight.device == metric_device, f"{type(re._weight.device)}:{re._weight.device} vs "
                f"{type(metric_device)}:{metric_device}"

    @pytest.mark.parametrize("average", [False, "macro", "weighted", "micro", "samples"])
    def test_multilabel_accumulator_device(self, average):
        # Multiclass input data of shape (N, ) and (N, C)

        metric_devices = [torch.device("cpu")]
        device = idist.device()
        if device.type != "xla":
            metric_devices.append(idist.device())

        for metric_device in metric_devices:
            re = Recall(is_multilabel=True, average=average, device=metric_device)
            assert re._device == metric_device
            assert re._updated is False

            y_reed = torch.randint(0, 2, size=(10, 4, 20, 23))
            y = torch.randint(0, 2, size=(10, 4, 20, 23)).long()
            re.update((y_reed, y))

            assert re._updated is True

            assert (
                re._numerator.device == metric_device
            ), f"{type(re._numerator.device)}:{re._numerator.device} vs {type(metric_device)}:{metric_device}"

            if average != "samples":
                # For average='samples', `_denominator` is of type `int` so it has not `device` member.
                assert (
                    re._denominator.device == metric_device
                ), f"{type(re._denominator.device)}:{re._denominator.device} vs {type(metric_device)}:{metric_device}"

            if average == "weighted":
                assert re._weight.device == metric_device, f"{type(re._weight.device)}:{re._weight.device} vs "
                f"{type(metric_device)}:{metric_device}"



"""
 MNIST example with training and validation monitoring using Neptune.

 Requirements:
    Neptune: `pip install neptune`

 Usage:

    Run the example:
    ```bash
    python mnist_with_neptune_logger.py
    ```

    Go to https://neptune.ai and explore your run.

Note:
    You can view example runs here:
    https://app.neptune.ai/o/common/org/pytorch-ignite-integration/
"""

from argparse import ArgumentParser

import torch
import torch.nn.functional as F
from torch import nn
from torch.optim import SGD
from torch.utils.data import DataLoader
from torchvision.datasets import MNIST
from torchvision.transforms import Compose, Normalize, ToTensor

from ignite.engine import create_supervised_evaluator, create_supervised_trainer, Events
from ignite.handlers import Checkpoint

from ignite.handlers.neptune_logger import (
    global_step_from_engine,
    GradsScalarHandler,
    NeptuneLogger,
    NeptuneSaver,
    WeightsScalarHandler,
)
from ignite.metrics import Accuracy, Loss
from ignite.utils import setup_logger


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=-1)


def get_data_loaders(train_batch_size, val_batch_size):
    data_transform = Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])

    train_loader = DataLoader(
        MNIST(download=True, root=".", transform=data_transform, train=True), batch_size=train_batch_size, shuffle=True
    )

    val_loader = DataLoader(
        MNIST(download=False, root=".", transform=data_transform, train=False), batch_size=val_batch_size, shuffle=False
    )
    return train_loader, val_loader


def run(train_batch_size, val_batch_size, epochs, lr, momentum):
    train_loader, val_loader = get_data_loaders(train_batch_size, val_batch_size)
    model = Net()
    device = "cpu"

    if torch.cuda.is_available():
        device = "cuda"

    model.to(device)  # Move model before creating optimizer
    optimizer = SGD(model.parameters(), lr=lr, momentum=momentum)
    criterion = nn.CrossEntropyLoss()
    trainer = create_supervised_trainer(model, optimizer, criterion, device=device)
    trainer.logger = setup_logger("Trainer")

    metrics = {"accuracy": Accuracy(), "loss": Loss(criterion)}

    train_evaluator = create_supervised_evaluator(model, metrics=metrics, device=device)
    train_evaluator.logger = setup_logger("Train Evaluator")
    validation_evaluator = create_supervised_evaluator(model, metrics=metrics, device=device)
    validation_evaluator.logger = setup_logger("Val Evaluator")

    @trainer.on(Events.EPOCH_COMPLETED)
    def compute_metrics(engine):
        train_evaluator.run(train_loader)
        validation_evaluator.run(val_loader)

    npt_logger = NeptuneLogger(
        api_token="ANONYMOUS",
        project="common/pytorch-ignite-integration",
        name="ignite-mnist-example",
    )

    npt_logger.experiment["params"] = {
        "train_batch_size": train_batch_size,
        "val_batch_size": val_batch_size,
        "epochs": epochs,
        "lr": lr,
        "momentum": momentum,
    }

    npt_logger.attach_output_handler(
        trainer,
        event_name=Events.ITERATION_COMPLETED(every=100),
        tag="training",
        output_transform=lambda loss: {"batchloss": loss},
    )

    for tag, evaluator in [("training", train_evaluator), ("validation", validation_evaluator)]:
        npt_logger.attach_output_handler(
            evaluator,
            event_name=Events.EPOCH_COMPLETED,
            tag=tag,
            metric_names=["loss", "accuracy"],
            global_step_transform=global_step_from_engine(trainer),
        )

    npt_logger.attach_opt_params_handler(trainer, event_name=Events.ITERATION_COMPLETED(every=100), optimizer=optimizer)

    npt_logger.attach(
        trainer, log_handler=WeightsScalarHandler(model), event_name=Events.ITERATION_COMPLETED(every=100)
    )

    npt_logger.attach(trainer, log_handler=GradsScalarHandler(model), event_name=Events.ITERATION_COMPLETED(every=100))

    def score_function(engine):
        return engine.state.metrics["accuracy"]

    handler = Checkpoint(
        {"model": model},
        NeptuneSaver(npt_logger),
        n_saved=2,
        filename_prefix="best",
        score_function=score_function,
        score_name="validation_accuracy",
        global_step_transform=global_step_from_engine(trainer),
    )
    validation_evaluator.add_event_handler(Events.COMPLETED, handler)

    # kick everything off
    trainer.run(train_loader, max_epochs=epochs)

    npt_logger.close()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--batch_size", type=int, default=64, help="input batch size for training (default: 64)")
    parser.add_argument(
        "--val_batch_size", type=int, default=1000, help="input batch size for validation (default: 1000)"
    )
    parser.add_argument("--epochs", type=int, default=10, help="number of epochs to train (default: 10)")
    parser.add_argument("--lr", type=float, default=0.01, help="learning rate (default: 0.01)")
    parser.add_argument("--momentum", type=float, default=0.5, help="SGD momentum (default: 0.5)")

    args = parser.parse_args()

    run(args.batch_size, args.val_batch_size, args.epochs, args.lr, args.momentum)



import pytest
import torch

from ignite import distributed as idist
from ignite.engine import Engine
from ignite.metrics import Accuracy, MetricGroup, Precision

torch.manual_seed(41)


def test_update():
    precision = Precision()
    accuracy = Accuracy()

    group = MetricGroup({"precision": Precision(), "accuracy": Accuracy()})

    y_pred = torch.randint(0, 2, (100,))
    y = torch.randint(0, 2, (100,))

    precision.update((y_pred, y))
    accuracy.update((y_pred, y))
    group.update((y_pred, y))

    assert precision.state_dict() == group.metrics["precision"].state_dict()
    assert accuracy.state_dict() == group.metrics["accuracy"].state_dict()


def test_output_transform():
    def drop_first(output):
        y_pred, y = output
        return (y_pred[1:], y[1:])

    precision = Precision(output_transform=drop_first)
    accuracy = Accuracy(output_transform=drop_first)

    group = MetricGroup(
        {"precision": Precision(output_transform=drop_first), "accuracy": Accuracy(output_transform=drop_first)}
    )

    y_pred = torch.randint(0, 2, (100,))
    y = torch.randint(0, 2, (100,))

    precision.update(drop_first(drop_first((y_pred, y))))
    accuracy.update(drop_first(drop_first((y_pred, y))))
    group.update(drop_first((y_pred, y)))

    assert precision.state_dict() == group.metrics["precision"].state_dict()
    assert accuracy.state_dict() == group.metrics["accuracy"].state_dict()


def test_compute():
    precision = Precision()
    accuracy = Accuracy()

    group = MetricGroup({"precision": Precision(), "accuracy": Accuracy()})

    for _ in range(3):
        y_pred = torch.randint(0, 2, (100,))
        y = torch.randint(0, 2, (100,))

        precision.update((y_pred, y))
        accuracy.update((y_pred, y))
        group.update((y_pred, y))

    assert group.compute() == {"precision": precision.compute(), "accuracy": accuracy.compute()}

    precision.reset()
    accuracy.reset()
    group.reset()

    assert precision.state_dict() == group.metrics["precision"].state_dict()
    assert accuracy.state_dict() == group.metrics["accuracy"].state_dict()


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_integration(self):
        rank = idist.get_rank()
        torch.manual_seed(12 + rank)

        n_epochs = 3
        n_iters = 5
        batch_size = 10
        device = idist.device()

        y_true = torch.randint(0, 2, size=(n_iters * batch_size,)).to(device)
        y_pred = torch.randint(0, 2, (n_iters * batch_size,)).to(device)

        def update(_, i):
            return (
                y_pred[i * batch_size : (i + 1) * batch_size],
                y_true[i * batch_size : (i + 1) * batch_size],
            )

        engine = Engine(update)

        precision = Precision()
        precision.attach(engine, "precision")

        accuracy = Accuracy()
        accuracy.attach(engine, "accuracy")

        group = MetricGroup({"eval_metrics.accuracy": Accuracy(), "eval_metrics.precision": Precision()})
        group.attach(engine, "eval_metrics")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=n_epochs)

        assert "eval_metrics" in engine.state.metrics
        assert "eval_metrics.accuracy" in engine.state.metrics
        assert "eval_metrics.precision" in engine.state.metrics

        assert engine.state.metrics["eval_metrics"] == {
            "eval_metrics.accuracy": engine.state.metrics["accuracy"],
            "eval_metrics.precision": engine.state.metrics["precision"],
        }
        assert engine.state.metrics["eval_metrics.accuracy"] == engine.state.metrics["accuracy"]
        assert engine.state.metrics["eval_metrics.precision"] == engine.state.metrics["precision"]



from unittest.mock import Mock, patch

import pytest
import torch

from ignite.engine import Engine, State

from ignite.metrics import GpuInfo


def test_no_pynvml_package():
    with patch.dict("sys.modules", {"pynvml.smi": None}):
        with pytest.raises(ModuleNotFoundError, match="This contrib module requires pynvml to be installed."):
            GpuInfo()


@pytest.mark.skipif(torch.cuda.is_available(), reason="Skip if GPU")
def test_no_gpu():
    with pytest.raises(RuntimeError, match="This contrib module requires available GPU"):
        GpuInfo()


def _test_gpu_info(device="cpu"):
    gpu_info = GpuInfo()

    # increase code cov
    gpu_info.reset()
    gpu_info.update(None)

    t = torch.rand(4, 10, 100, 100).to(device)
    data = gpu_info.compute()
    assert len(data) > 0
    assert "fb_memory_usage" in data[0]
    mem_report = data[0]["fb_memory_usage"]
    assert "used" in mem_report and "total" in mem_report
    assert mem_report["total"] > 0.0
    assert mem_report["used"] > t.shape[0] * t.shape[1] * t.shape[2] * t.shape[3] / 1024.0 / 1024.0

    assert "utilization" in data[0]
    util_report = data[0]["utilization"]
    assert "gpu_util" in util_report

    # with Engine
    engine = Engine(lambda engine, batch: 0.0)
    engine.state = State(metrics={})

    gpu_info.completed(engine, name="gpu")

    assert "gpu:0 mem(%)" in engine.state.metrics

    assert isinstance(engine.state.metrics["gpu:0 mem(%)"], int)
    assert int(mem_report["used"] * 100.0 / mem_report["total"]) == engine.state.metrics["gpu:0 mem(%)"]

    if util_report["gpu_util"] != "N/A":
        assert "gpu:0 util(%)" in engine.state.metrics
        assert isinstance(engine.state.metrics["gpu:0 util(%)"], int)
        assert int(util_report["gpu_util"]) == engine.state.metrics["gpu:0 util(%)"]
    else:
        assert "gpu:0 util(%)" not in engine.state.metrics


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_gpu_info_on_cuda():
    _test_gpu_info(device="cuda")


query_resp = None


@pytest.fixture
def mock_pynvml_module():
    with patch.dict(
        "sys.modules",
        {
            "pynvml": Mock(name="pynvml"),
            "pynvml.smi": Mock(name="pynvml.smi"),
            "pynvml.smi.nvidia_smi": Mock(name="pynvml.smi.nvidia_smi"),
        },
    ):
        import pynvml
        from pynvml.smi import nvidia_smi

        def query(*args, **kwargs):
            return query_resp

        def getInstance():
            nvsmi = Mock()
            nvsmi.DeviceQuery = Mock(side_effect=query)
            return nvsmi

        nvidia_smi.getInstance = Mock(side_effect=getInstance)
        yield pynvml


@pytest.fixture
def mock_gpu_is_available():
    with patch("ignite.metrics.gpu_info.torch.cuda") as mock_cuda:
        mock_cuda.is_available.return_value = True
        yield mock_cuda


@pytest.mark.skipif(torch.cuda.is_available(), reason="No need to mock if has GPU")
def test_gpu_info_mock(mock_pynvml_module, mock_gpu_is_available):
    global query_resp

    query_resp = {"gpu": [{"fb_memory_usage": {"used": 100.0, "total": 11000.0}, "utilization": {"gpu_util": 50.0}}]}

    assert torch.cuda.is_available()
    _test_gpu_info()

    # Tests https://github.com/pytorch/ignite/issues/1040
    query_resp = {"gpu": [{"fb_memory_usage": {"used": 100.0, "total": 11000.0}, "utilization": {"gpu_util": "N/A"}}]}
    _test_gpu_info()

    def _test_with_custom_query(resp, warn_msg, check_compute=False):
        from pynvml.smi import nvidia_smi

        def query(*args, **kwargs):
            return resp

        def getInstance():
            nvsmi = Mock()
            nvsmi.DeviceQuery = Mock(side_effect=query)
            return nvsmi

        nvidia_smi.getInstance = Mock(side_effect=getInstance)
        gpu_info = GpuInfo()
        if check_compute:
            with pytest.warns(UserWarning, match=warn_msg):
                gpu_info.compute()

        # with Engine
        engine = Engine(lambda engine, batch: 0.0)
        engine.state = State(metrics={})

        with pytest.warns(UserWarning, match=warn_msg):
            gpu_info.completed(engine, name="gpu info")

    # No GPU info
    _test_with_custom_query(resp={}, warn_msg=r"No GPU information available", check_compute=True)

    # No GPU memory info
    _test_with_custom_query(resp={"gpu": [{"utilization": {}}]}, warn_msg=r"No GPU memory usage information available")

    # No GPU utilization info
    _test_with_custom_query(
        resp={"gpu": [{"fb_memory_usage": {}}]}, warn_msg=r"No GPU utilization information available"
    )



# coding: utf-8

import unittest.mock as mock

import pytest
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from ignite.contrib.engines import create_supervised_tbptt_trainer, Tbptt_Events
from ignite.contrib.engines.tbptt import _detach_hidden


def test_detach_hidden_RNN():
    # Create hidden vector (in tuple)
    X = torch.ones(2, 3, 4)
    model = nn.RNN(4, 1)
    _, hidden = model(X)

    # Function to test
    hidden_ = _detach_hidden(hidden)

    assert hidden_.grad_fn is None  # properly detached
    assert (hidden == hidden_).all().item() == 1  # Equal values


def test_detach_hidden_LSTM():
    # Create hidden vector (in tuple)
    X = torch.ones(2, 3, 4)
    model = nn.LSTM(4, 1)
    _, hidden = model(X)

    # Function to test
    hidden_ = _detach_hidden(hidden)

    for h, h_ in zip(hidden, hidden_):
        assert h_.grad_fn is None  # properly detached
        assert (h == h_).all().item() == 1  # Equal values


def test_detach_hidden_raise():
    with pytest.raises(TypeError):
        _detach_hidden(0)


@mock.patch("ignite.contrib.engines.tbptt._detach_hidden")
def test_create_supervised_tbptt_trainer_callcounts(mock_detach_hidden):
    # Mocking objects
    model = mock.MagicMock()
    # Necessary to unpack output
    model.return_value = (1, 1)
    optimizer = mock.MagicMock()
    loss = mock.MagicMock()

    trainer = create_supervised_tbptt_trainer(model, optimizer, loss, tbtt_step=2)

    # Adding two mock handles to the trainer to monitor that TBPTT events are
    # called correctly
    handle_started = mock.MagicMock()
    trainer.add_event_handler(Tbptt_Events.TIME_ITERATION_STARTED, handle_started)
    handle_completed = mock.MagicMock()
    trainer.add_event_handler(Tbptt_Events.TIME_ITERATION_COMPLETED, handle_completed)

    # Fake data
    X = torch.ones(6, 2, 1)
    y = torch.ones(6, 2, 1)
    data = [(X, y)]

    # Running trainer
    trainer.run(data)

    # Verifications
    assert handle_started.call_count == 3
    assert handle_completed.call_count == 3
    assert mock_detach_hidden.call_count == 2
    assert model.call_count == 3
    assert loss.call_count == 3
    assert optimizer.zero_grad.call_count == 3
    assert optimizer.step.call_count == 3
    n_args_tuple = tuple(len(args) for args, kwargs in model.call_args_list)
    assert n_args_tuple == (1, 2, 2)


def _test_create_supervised_tbptt_trainer(device):
    # Defining dummy recurrent model with zero weights
    model = nn.RNN(1, 1, bias=False)
    model.to(device)  # Move model before creating optimizer
    for p in model.parameters():
        p.data.zero_()

    # Set some mock on forward to monitor
    forward_mock = mock.MagicMock()
    forward_mock.return_value = None
    model.register_forward_hook(forward_mock)

    # Defning optimizer and trainer
    optimizer = optim.SGD(model.parameters(), 1)
    trainer = create_supervised_tbptt_trainer(model, optimizer, F.mse_loss, tbtt_step=2, device=device)

    # Fake data
    X = torch.ones(6, 2, 1)
    y = torch.ones(6, 2, 1)
    data = [(X, y)]

    # Running trainer
    trainer.run(data)

    # If tbptt is not use (one gradient update), the hidden to hidden weight
    # should stay zero
    assert not model.weight_hh_l0.item() == pytest.approx(0)

    # Cheking forward calls
    assert forward_mock.call_count == 3
    for i in range(3):
        inputs = forward_mock.call_args_list[i][0][1]
        if i == 0:
            assert len(inputs) == 1
        else:
            assert len(inputs) == 2
            x, h = inputs
            assert h.is_leaf


def test_create_supervised_tbptt_trainer_with_cpu():
    _test_create_supervised_tbptt_trainer("cpu")


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Skip if no GPU")
def test_create_supervised_tbptt_trainer_on_cuda():
    _test_create_supervised_tbptt_trainer("cuda")



import re
from pathlib import Path
from unittest.mock import patch

import pytest
import torch
import torch.nn as nn
from packaging.version import Version

from ignite.engine import Engine, Events
from ignite.handlers.state_param_scheduler import (
    ExpStateScheduler,
    LambdaStateScheduler,
    MultiStepStateScheduler,
    PiecewiseLinearStateScheduler,
    StepStateScheduler,
)

config1 = (3, [(2, 0), (5, 10)], True, [0.0, 0.0, 3.3333333333333335])
expected_hist2 = [0.0] * 10 + [float(i) for i in range(1, 11)] + [10.0] * 10
config2 = (30, [(10, 0), (20, 10)], True, expected_hist2)
config3 = (
    PiecewiseLinearStateScheduler,
    {"param_name": "linear_scheduled_param", "milestones_values": [(3, 12), (5, 10)], "create_new": True},
)
config4 = (
    ExpStateScheduler,
    {"param_name": "exp_scheduled_param", "initial_value": 10, "gamma": 0.99, "create_new": True},
)
config5 = (
    MultiStepStateScheduler,
    {
        "param_name": "multistep_scheduled_param",
        "initial_value": 10,
        "gamma": 0.99,
        "milestones": [3, 6],
        "create_new": True,
    },
)

if Version(torch.__version__) < Version("1.9.0"):
    torch_testing_assert_close = torch.testing.assert_allclose
else:
    torch_testing_assert_close = torch.testing.assert_close


class LambdaState:
    def __init__(self, initial_value, gamma):
        self.initial_value = initial_value
        self.gamma = gamma

    def __call__(self, event_index):
        return self.initial_value * self.gamma ** (event_index % 9)


config6 = (
    LambdaStateScheduler,
    {
        "param_name": "custom_scheduled_param",
        "lambda_obj": LambdaState(initial_value=10, gamma=0.99),
        "create_new": True,
    },
)

config7 = (
    StepStateScheduler,
    {"param_name": "step_scheduled_param", "initial_value": 10, "gamma": 0.99, "step_size": 5, "create_new": True},
)


@pytest.mark.parametrize("max_epochs, milestones_values,  save_history, expected_param_history", [config1, config2])
def test_pwlinear_scheduler_linear_increase_history(
    max_epochs, milestones_values, save_history, expected_param_history
):
    # Testing linear increase
    engine = Engine(lambda e, b: None)
    pw_linear_step_parameter_scheduler = PiecewiseLinearStateScheduler(
        param_name="pwlinear_scheduled_param",
        milestones_values=milestones_values,
        save_history=save_history,
        create_new=True,
    )
    pw_linear_step_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    expected_param_history = expected_param_history
    assert hasattr(engine.state, "param_history")
    state_param = engine.state.param_history["pwlinear_scheduled_param"]
    assert len(state_param) == len(expected_param_history)
    assert state_param == expected_param_history

    state_dict = pw_linear_step_parameter_scheduler.state_dict()
    pw_linear_step_parameter_scheduler.load_state_dict(state_dict)


@pytest.mark.parametrize("max_epochs, milestones_values", [(3, [(3, 12), (5, 10)]), (5, [(10, 12), (20, 10)])])
def test_pwlinear_scheduler_step_constant(max_epochs, milestones_values):
    # Testing step_constant
    engine = Engine(lambda e, b: None)
    linear_state_parameter_scheduler = PiecewiseLinearStateScheduler(
        param_name="pwlinear_scheduled_param", milestones_values=milestones_values, create_new=True
    )
    linear_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(getattr(engine.state, "pwlinear_scheduled_param"), float(milestones_values[0][1]))

    state_dict = linear_state_parameter_scheduler.state_dict()
    linear_state_parameter_scheduler.load_state_dict(state_dict)


@pytest.mark.parametrize(
    "max_epochs, milestones_values, expected_val",
    [(2, [(0, 0), (3, 10)], 6.666666666666667), (10, [(0, 0), (20, 10)], 5.0)],
)
def test_pwlinear_scheduler_linear_increase(max_epochs, milestones_values, expected_val):
    # Testing linear increase
    engine = Engine(lambda e, b: None)
    linear_state_parameter_scheduler = PiecewiseLinearStateScheduler(
        param_name="pwlinear_scheduled_param", milestones_values=milestones_values, create_new=True
    )
    linear_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(getattr(engine.state, "pwlinear_scheduled_param"), expected_val, atol=0.001, rtol=0.0)

    state_dict = linear_state_parameter_scheduler.state_dict()
    linear_state_parameter_scheduler.load_state_dict(state_dict)


@pytest.mark.parametrize("max_epochs, milestones_values,", [(3, [(0, 0), (3, 10)]), (40, [(0, 0), (20, 10)])])
def test_pwlinear_scheduler_max_value(max_epochs, milestones_values):
    # Testing max_value
    engine = Engine(lambda e, b: None)
    linear_state_parameter_scheduler = PiecewiseLinearStateScheduler(
        param_name="linear_scheduled_param", milestones_values=milestones_values, create_new=True
    )
    linear_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(getattr(engine.state, "linear_scheduled_param"), float(milestones_values[-1][1]))

    state_dict = linear_state_parameter_scheduler.state_dict()
    linear_state_parameter_scheduler.load_state_dict(state_dict)


def test_piecewiselinear_asserts():
    with pytest.raises(TypeError, match=r"Argument milestones_values should be a list or tuple"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=None)

    with pytest.raises(ValueError, match=r"Argument milestones_values should be with at least one value"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=[])

    with pytest.raises(ValueError, match=r"Argument milestones_values should be a list of pairs"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=[(0.5,)])

    with pytest.raises(ValueError, match=r"Argument milestones_values should be a list of pairs"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=[(10, 0.5), (0.6,)])

    with pytest.raises(ValueError, match=r"Milestones should be increasing integers"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=[(10, 0.5), (5, 0.6)])

    with pytest.raises(TypeError, match=r"Value of a milestone should be integer"):
        PiecewiseLinearStateScheduler(param_name="linear_scheduled_param", milestones_values=[(0.5, 1)])


@pytest.mark.parametrize("max_epochs, initial_value, gamma", [(3, 10, 0.99), (40, 5, 0.98)])
def test_exponential_scheduler(max_epochs, initial_value, gamma):
    engine = Engine(lambda e, b: None)
    exp_state_parameter_scheduler = ExpStateScheduler(
        param_name="exp_scheduled_param", initial_value=initial_value, gamma=gamma, create_new=True
    )
    exp_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(getattr(engine.state, "exp_scheduled_param"), initial_value * gamma**max_epochs)

    state_dict = exp_state_parameter_scheduler.state_dict()
    exp_state_parameter_scheduler.load_state_dict(state_dict)


@pytest.mark.parametrize("max_epochs, initial_value, gamma, step_size", [(3, 10, 0.99, 5), (40, 5, 0.98, 22)])
def test_step_scheduler(max_epochs, initial_value, gamma, step_size):
    engine = Engine(lambda e, b: None)
    step_state_parameter_scheduler = StepStateScheduler(
        param_name="step_scheduled_param",
        initial_value=initial_value,
        gamma=gamma,
        step_size=step_size,
        create_new=True,
    )
    step_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(
        getattr(engine.state, "step_scheduled_param"), initial_value * gamma ** (max_epochs // step_size)
    )

    state_dict = step_state_parameter_scheduler.state_dict()
    step_state_parameter_scheduler.load_state_dict(state_dict)


from bisect import bisect_right


@pytest.mark.parametrize(
    "max_epochs, initial_value, gamma, milestones", [(3, 10, 0.99, [3, 6]), (40, 5, 0.98, [3, 6, 9, 10, 11])]
)
def test_multistep_scheduler(max_epochs, initial_value, gamma, milestones):
    engine = Engine(lambda e, b: None)
    multi_step_state_parameter_scheduler = MultiStepStateScheduler(
        param_name="multistep_scheduled_param",
        initial_value=initial_value,
        gamma=gamma,
        milestones=milestones,
        create_new=True,
    )
    multi_step_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=max_epochs)
    torch_testing_assert_close(
        getattr(engine.state, "multistep_scheduled_param"),
        initial_value * gamma ** bisect_right(milestones, max_epochs),
    )

    state_dict = multi_step_state_parameter_scheduler.state_dict()
    multi_step_state_parameter_scheduler.load_state_dict(state_dict)


def test_custom_scheduler():
    engine = Engine(lambda e, b: None)

    class LambdaState:
        def __init__(self, initial_value, gamma):
            self.initial_value = initial_value
            self.gamma = gamma

        def __call__(self, event_index):
            return self.initial_value * self.gamma ** (event_index % 9)

    lambda_state_parameter_scheduler = LambdaStateScheduler(
        param_name="custom_scheduled_param", lambda_obj=LambdaState(initial_value=10, gamma=0.99), create_new=True
    )
    lambda_state_parameter_scheduler.attach(engine, Events.EPOCH_COMPLETED)
    engine.run([0] * 8, max_epochs=2)
    torch_testing_assert_close(
        getattr(engine.state, "custom_scheduled_param"), LambdaState(initial_value=10, gamma=0.99)(2)
    )
    engine.run([0] * 8, max_epochs=20)
    torch_testing_assert_close(
        getattr(engine.state, "custom_scheduled_param"), LambdaState(initial_value=10, gamma=0.99)(20)
    )

    state_dict = lambda_state_parameter_scheduler.state_dict()
    lambda_state_parameter_scheduler.load_state_dict(state_dict)


def test_custom_scheduler_asserts():
    class LambdaState:
        def __init__(self, initial_value, gamma):
            self.initial_value = initial_value
            self.gamma = gamma

    with pytest.raises(ValueError, match=r"Expected lambda_obj to be callable."):
        lambda_state_parameter_scheduler = LambdaStateScheduler(
            param_name="custom_scheduled_param", lambda_obj=LambdaState(initial_value=10, gamma=0.99), create_new=True
        )


@pytest.mark.parametrize("scheduler_cls, scheduler_kwargs", [config3, config4, config5, config6])
def test_simulate_and_plot_values(scheduler_cls, scheduler_kwargs):
    import matplotlib

    matplotlib.use("Agg")

    event = Events.EPOCH_COMPLETED
    max_epochs = 2
    data = [0] * 10

    scheduler = scheduler_cls(**scheduler_kwargs)
    trainer = Engine(lambda engine, batch: None)
    scheduler.attach(trainer, event)
    trainer.run(data, max_epochs=max_epochs)

    # launch plot values
    scheduler_cls.plot_values(num_events=len(data) * max_epochs, **scheduler_kwargs)


@pytest.mark.parametrize("save_history", [False, True])
@pytest.mark.parametrize("scheduler_cls, scheduler_kwargs", [config3, config4, config5, config6])
def test_simulate_values(scheduler_cls, scheduler_kwargs, save_history):
    max_epochs = 2
    data = [0] * 10
    scheduler_kwargs["save_history"] = save_history
    scheduler_cls.simulate_values(num_events=len(data) * max_epochs, **scheduler_kwargs)


def test_torch_save_load(dirname):
    lambda_state_parameter_scheduler = LambdaStateScheduler(
        param_name="custom_scheduled_param", lambda_obj=LambdaState(initial_value=10, gamma=0.99), create_new=True
    )

    filepath = Path(dirname) / "dummy_lambda_state_parameter_scheduler.pt"
    torch.save(lambda_state_parameter_scheduler, filepath)
    if Version(torch.__version__) >= Version("1.13.0"):
        kwargs = {"weights_only": False}
    else:
        kwargs = {}
    loaded_lambda_state_parameter_scheduler = torch.load(filepath, **kwargs)

    engine1 = Engine(lambda e, b: None)
    lambda_state_parameter_scheduler.attach(engine1, Events.EPOCH_COMPLETED)
    engine1.run([0] * 8, max_epochs=2)
    torch_testing_assert_close(
        getattr(engine1.state, "custom_scheduled_param"), LambdaState(initial_value=10, gamma=0.99)(2)
    )

    engine2 = Engine(lambda e, b: None)
    loaded_lambda_state_parameter_scheduler.attach(engine2, Events.EPOCH_COMPLETED)
    engine2.run([0] * 8, max_epochs=2)
    torch_testing_assert_close(
        getattr(engine2.state, "custom_scheduled_param"), LambdaState(initial_value=10, gamma=0.99)(2)
    )
    torch_testing_assert_close(
        getattr(engine1.state, "custom_scheduled_param"), getattr(engine2.state, "custom_scheduled_param")
    )


def test_simulate_and_plot_values_no_matplotlib():
    with pytest.raises(ModuleNotFoundError, match=r"This method requires matplotlib to be installed."):
        with patch.dict("sys.modules", {"matplotlib.pyplot": None}):
            event = Events.EPOCH_COMPLETED
            max_epochs = 2
            data = [0] * 10

            kwargs = {
                "param_name": "multistep_scheduled_param",
                "initial_value": 10,
                "gamma": 0.99,
                "milestones": [3, 6],
                "create_new": True,
            }
            scheduler = MultiStepStateScheduler(**kwargs)
            trainer = Engine(lambda engine, batch: None)
            scheduler.attach(trainer, event)
            trainer.run(data, max_epochs=max_epochs)

            # launch plot values
            MultiStepStateScheduler.plot_values(num_events=len(data) * max_epochs, **kwargs)


def test_multiple_scheduler_with_save_history():
    engine_multiple_schedulers = Engine(lambda e, b: None)
    configs = [config3, config4, config5, config6, config7]
    for scheduler, config in configs:
        if "save_history" in config:
            del config["save_history"]
        _scheduler = scheduler(**config, save_history=True)
        _scheduler.attach(engine_multiple_schedulers)

    engine_multiple_schedulers.run([0] * 8, max_epochs=2)

    for scheduler, config in configs:
        engine = Engine(lambda e, b: None)
        _scheduler = scheduler(**config, save_history=True)
        _scheduler.attach(engine)
        engine.run([0] * 8, max_epochs=2)
        torch_testing_assert_close(
            engine_multiple_schedulers.state.param_history[config["param_name"]],
            engine.state.param_history[config["param_name"]],
        )


def test_docstring_examples():
    # LambdaStateScheduler

    engine = Engine(lambda e, b: None)

    class LambdaState:
        def __init__(self, initial_value, gamma):
            self.initial_value = initial_value
            self.gamma = gamma

        def __call__(self, event_index):
            return self.initial_value * self.gamma ** (event_index % 9)

    param_scheduler = LambdaStateScheduler(param_name="param", lambda_obj=LambdaState(10, 0.99), create_new=True)

    param_scheduler.attach(engine, Events.EPOCH_COMPLETED)

    engine.run([0] * 8, max_epochs=2)

    # PiecewiseLinearStateScheduler

    engine = Engine(lambda e, b: None)

    param_scheduler = PiecewiseLinearStateScheduler(
        param_name="param", milestones_values=[(10, 0.5), (20, 0.45), (21, 0.3), (30, 0.1), (40, 0.1)], create_new=True
    )

    param_scheduler.attach(engine, Events.EPOCH_COMPLETED)

    engine.run([0] * 8, max_epochs=40)

    # ExpStateScheduler

    engine = Engine(lambda e, b: None)

    param_scheduler = ExpStateScheduler(param_name="param", initial_value=10, gamma=0.99, create_new=True)

    param_scheduler.attach(engine, Events.EPOCH_COMPLETED)

    engine.run([0] * 8, max_epochs=2)

    # StepStateScheduler

    engine = Engine(lambda e, b: None)

    param_scheduler = StepStateScheduler(param_name="param", initial_value=10, gamma=0.99, step_size=5, create_new=True)

    param_scheduler.attach(engine, Events.EPOCH_COMPLETED)

    engine.run([0] * 8, max_epochs=10)

    # MultiStepStateScheduler

    engine = Engine(lambda e, b: None)

    param_scheduler = MultiStepStateScheduler(
        param_name="param", initial_value=10, gamma=0.99, milestones=[3, 6], create_new=True
    )

    param_scheduler.attach(engine, Events.EPOCH_COMPLETED)

    engine.run([0] * 8, max_epochs=10)


def test_param_scheduler_attach_exception():
    trainer = Engine(lambda e, b: None)
    param_name = "state_param"

    setattr(trainer.state, param_name, None)

    save_history = True
    create_new = True

    param_scheduler = PiecewiseLinearStateScheduler(
        param_name=param_name,
        milestones_values=[(0, 0.0), (10, 0.999)],
        save_history=save_history,
        create_new=create_new,
    )

    with pytest.raises(
        ValueError,
        match=r"Attribute '" + re.escape(param_name) + "' already exists in the engine.state. "
        r"This may be a conflict between multiple handlers. "
        r"Please choose another name.",
    ):
        param_scheduler.attach(trainer, Events.ITERATION_COMPLETED)


def test_param_scheduler_attach_warning():
    trainer = Engine(lambda e, b: None)
    param_name = "state_param"
    save_history = True
    create_new = False

    param_scheduler = PiecewiseLinearStateScheduler(
        param_name=param_name,
        milestones_values=[(0, 0.0), (10, 0.999)],
        save_history=save_history,
        create_new=create_new,
    )

    with pytest.warns(
        UserWarning,
        match=r"Attribute '" + re.escape(param_name) + "' is not defined in the engine.state. "
        r"PiecewiseLinearStateScheduler will create it. Remove this warning by setting create_new=True.",
    ):
        param_scheduler.attach(trainer, Events.ITERATION_COMPLETED)


def test_param_scheduler_with_ema_handler():
    from ignite.handlers import EMAHandler

    model = nn.Linear(2, 1)
    trainer = Engine(lambda e, b: model(b))
    data = torch.rand(100, 2)

    param_name = "ema_decay"

    ema_handler = EMAHandler(model)
    ema_handler.attach(trainer, name=param_name, event=Events.ITERATION_COMPLETED)

    ema_decay_scheduler = PiecewiseLinearStateScheduler(
        param_name=param_name, milestones_values=[(0, 0.0), (10, 0.999)], save_history=True
    )
    ema_decay_scheduler.attach(trainer, Events.ITERATION_COMPLETED)
    trainer.run(data, max_epochs=20)



# Basic training configuration
import os

import albumentations as A
import cv2
from albumentations.pytorch import ToTensorV2 as ToTensor
from dataflow import get_inference_dataloader, ignore_mask_boundaries
from torchvision.models.segmentation import deeplabv3_resnet101

# ##############################
# Global configs
# ##############################

seed = 21
device = "cuda"
debug = False
# Use AMP with torch native
with_amp = True


num_classes = 21
batch_size = 9  # total batch size
num_workers = 8  # total num workers per node

val_img_size = 513

# ##############################
# Setup Dataflow
# ##############################

assert "DATASET_PATH" in os.environ
data_path = os.environ["DATASET_PATH"]

assert "SBD_DATASET_PATH" in os.environ
sbd_data_path = os.environ["SBD_DATASET_PATH"]

mean = (0.485, 0.456, 0.406)
std = (0.229, 0.224, 0.225)


val_transforms = A.Compose(
    [
        A.PadIfNeeded(val_img_size, val_img_size, border_mode=cv2.BORDER_CONSTANT),
        A.Normalize(mean=mean, std=std),
        ignore_mask_boundaries,
        ToTensor(),
    ]
)


data_loader = get_inference_dataloader(
    root_path=data_path,
    mode="test",
    transforms=val_transforms,
    batch_size=batch_size,
    num_workers=num_workers,
    limit_num_samples=batch_size * 5 if debug else None,
)

# ##############################
# Setup model
# ##############################

num_classes = 21
model = deeplabv3_resnet101(num_classes=num_classes)


def model_output_transform(output):
    return output["out"]


# baseline_dplv3_resnet101_sbd: best_model_78_val_miou_bg=0.6871.pt
weights_path = "d8b4687d86cf445a944853fdd6a6b999"
# or can specify a path
# weights_path = "/path/to/best_model.pt"



import os
import sys
import time

import pytest
import torch

import ignite.distributed as idist
from ignite.engine import Engine, Events
from ignite.metrics import Frequency

if sys.platform.startswith("darwin"):
    pytest.skip("Skip if on MacOS", allow_module_level=True)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_nondistributed_average(available_device):
    artificial_time = 1  # seconds
    num_tokens = 100
    average_upper_bound = num_tokens / artificial_time
    average_lower_bound = average_upper_bound * 0.9
    freq_metric = Frequency(device=available_device)
    assert freq_metric._device == torch.device(available_device)
    freq_metric.reset()
    time.sleep(artificial_time)
    freq_metric.update(num_tokens)
    average = freq_metric.compute()
    assert average_lower_bound < average < average_upper_bound


def _test_frequency_with_engine(workers=None, lower_bound_factor=0.8, upper_bound_factor=1.1, every=1, device="cpu"):
    if workers is None:
        workers = idist.get_world_size()

    artificial_time = 1.0 / workers  # seconds
    total_tokens = 400 // workers
    batch_size = 128 // workers

    estimated_wps = batch_size * workers / artificial_time

    def update_fn(engine, batch):
        time.sleep(artificial_time)
        return {"ntokens": len(batch)}

    engine = Engine(update_fn)
    wps_metric = Frequency(output_transform=lambda x: x["ntokens"], device=device)
    assert wps_metric._device == torch.device(device)

    event = Events.ITERATION_COMPLETED(every=every)
    wps_metric.attach(engine, "wps", event_name=event)

    @engine.on(event)
    def assert_wps(e):
        wps = e.state.metrics["wps"]
        # Skip iterations 2, 3, 4 if backend is Horovod on CUDA,
        # wps is abnormally low for these iterations
        # otherwise, other values of wps are OK
        if idist.model_name() == "horovod-dist" and e.state.iteration in (2, 3, 4):
            return
        low_wps = estimated_wps * lower_bound_factor
        high_wps = estimated_wps * upper_bound_factor
        assert low_wps < wps <= high_wps, f"{e.state.iteration}: {low_wps} < {wps} <= {high_wps}"

    data = [[i] * batch_size for i in range(0, total_tokens, batch_size)]
    engine.run(data, max_epochs=2)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Skip on Windows")
def test_frequency_with_engine(available_device):
    _test_frequency_with_engine(workers=1, device=available_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_frequency_with_engine_distributed(distributed_context_single_node_gloo):
    _test_frequency_with_engine(workers=idist.get_world_size())


def test_frequency_with_engine_with_every(available_device):
    _test_frequency_with_engine(workers=1, every=1, device=available_device)
    _test_frequency_with_engine(workers=1, every=10, device=available_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_frequency_with_engine_distributed_with_every(distributed_context_single_node_gloo):
    _test_frequency_with_engine(workers=idist.get_world_size(), every=1)
    _test_frequency_with_engine(workers=idist.get_world_size(), every=10)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_frequency_with_engine, (None, 0.8, 1), np=nproc, do_init=True)
    gloo_hvd_executor(_test_frequency_with_engine, (None, 0.8, 10), np=nproc, do_init=True)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    _test_frequency_with_engine(workers=idist.get_world_size(), every=10)


def _test_distrib_xla_nprocs(index):
    _test_frequency_with_engine(workers=idist.get_world_size(), every=10)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



import argparse

import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
from model import Net
from torch.utils.data import DataLoader
from torchvision.transforms.functional import center_crop, resize, to_tensor

from ignite.engine import Engine, Events

from ignite.handlers import BasicTimeProfiler, ProgressBar
from ignite.metrics import PSNR

# Training settings
parser = argparse.ArgumentParser(description="PyTorch Super Res Example")
parser.add_argument("--crop_size", type=int, default=256, help="cropped size of the images for training")
parser.add_argument("--upscale_factor", type=int, required=True, help="super resolution upscale factor")
parser.add_argument("--batch_size", type=int, default=64, help="training batch size")
parser.add_argument("--test_batch_size", type=int, default=10, help="testing batch size")
parser.add_argument("--n_epochs", type=int, default=2, help="number of epochs to train for")
parser.add_argument("--lr", type=float, default=0.01, help="Learning Rate. Default=0.01")
parser.add_argument("--cuda", action="store_true", help="use cuda?")
parser.add_argument("--mps", action="store_true", default=False, help="enables macOS GPU training")
parser.add_argument("--threads", type=int, default=4, help="number of threads for data loader to use")
parser.add_argument("--seed", type=int, default=123, help="random seed to use. Default=123")
parser.add_argument("--debug", action="store_true", help="use debug")

opt = parser.parse_args()

print(opt)

if opt.cuda and not torch.cuda.is_available():
    raise Exception("No GPU found, please run without --cuda")
if not opt.mps and torch.backends.mps.is_available():
    raise Exception("Found mps device, please run with --mps to enable macOS GPU")

torch.manual_seed(opt.seed)
use_mps = opt.mps and torch.backends.mps.is_available()

if opt.cuda:
    device = torch.device("cuda")
elif use_mps:
    device = torch.device("mps")
else:
    device = torch.device("cpu")

print("===> Loading datasets")


class SRDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, scale_factor, crop_size=256):
        self.dataset = dataset
        self.scale_factor = scale_factor
        self.crop_size = crop_size

    def __getitem__(self, index):
        image, _ = self.dataset[index]
        img = image.convert("YCbCr")
        hr_image, _, _ = img.split()
        hr_image = center_crop(hr_image, self.crop_size)
        lr_image = hr_image.copy()
        if self.scale_factor != 1:
            size = self.crop_size // self.scale_factor
            lr_image = resize(lr_image, [size, size])
        hr_image = to_tensor(hr_image)
        lr_image = to_tensor(lr_image)
        return lr_image, hr_image

    def __len__(self):
        return len(self.dataset)


try:
    trainset = torchvision.datasets.Caltech101(root="./data", download=True)
    testset = torchvision.datasets.Caltech101(root="./data", download=False)
except RuntimeError:
    print("Dataset download problem, exiting without error code")
    exit(0)

trainset_sr = SRDataset(trainset, scale_factor=opt.upscale_factor, crop_size=opt.crop_size)
testset_sr = SRDataset(testset, scale_factor=opt.upscale_factor, crop_size=opt.crop_size)

training_data_loader = DataLoader(dataset=trainset_sr, num_workers=opt.threads, batch_size=opt.batch_size, shuffle=True)
testing_data_loader = DataLoader(dataset=testset_sr, num_workers=opt.threads, batch_size=opt.test_batch_size)

print("===> Building model")
model = Net(upscale_factor=opt.upscale_factor).to(device)
criterion = nn.MSELoss()

optimizer = optim.Adam(model.parameters(), lr=opt.lr)


def train_step(engine, batch):
    model.train()
    input, target = batch[0].to(device), batch[1].to(device)

    optimizer.zero_grad()
    loss = criterion(model(input), target)
    loss.backward()
    optimizer.step()

    return loss.item()


def validation_step(engine, batch):
    model.eval()
    with torch.no_grad():
        x, y = batch[0].to(device), batch[1].to(device)
        y_pred = model(x)

    return y_pred, y


trainer = Engine(train_step)
evaluator = Engine(validation_step)
psnr = PSNR(data_range=1)
psnr.attach(evaluator, "psnr")
validate_every = 1

if opt.debug:
    epoch_length = 10
    validate_epoch_length = 1
else:
    epoch_length = len(training_data_loader)
    validate_epoch_length = len(testing_data_loader)


@trainer.on(Events.EPOCH_COMPLETED(every=validate_every))
def log_validation():
    evaluator.run(testing_data_loader, epoch_length=validate_epoch_length)
    metrics = evaluator.state.metrics
    print(f"Epoch: {trainer.state.epoch}, Avg. PSNR: {metrics['psnr']} dB")


@trainer.on(Events.EPOCH_COMPLETED)
def checkpoint():
    model_out_path = "model_epoch_{}.pth".format(trainer.state.epoch)
    torch.save(model, model_out_path)
    print("Checkpoint saved to {}".format(model_out_path))


# Attach basic profiler
basic_profiler = BasicTimeProfiler()
basic_profiler.attach(trainer)

ProgressBar().attach(trainer, output_transform=lambda x: {"loss": x})

trainer.run(training_data_loader, opt.n_epochs, epoch_length=epoch_length)

results = basic_profiler.get_results()
basic_profiler.print_results(results)



import pytest
import torch

from ignite.distributed.comp_models.base import _SerialModel, _torch_version_gt_112, ComputationModel


def test_serial_model():
    _SerialModel.create_from_backend()
    model = _SerialModel.create_from_context()

    assert model.get_local_rank() == 0
    assert model.get_rank() == 0
    assert model.get_world_size() == 1
    assert model.get_nproc_per_node() == 1
    assert model.get_nnodes() == 1
    assert model.get_node_rank() == 0
    if torch.cuda.is_available():
        assert model.device().type == "cuda"
    elif _torch_version_gt_112 and torch.backends.mps.is_available():
        assert model.device().type == "mps"
    else:
        assert model.device().type == "cpu"
    assert model.backend() is None
    model.finalize()

    with pytest.raises(NotImplementedError, match=r"Serial computation model does not implement spawn method"):
        model.spawn()

    model.all_reduce(1)
    model.all_gather(1)
    model.broadcast(1)
    assert model._do_all_reduce(torch.tensor(1)) == torch.tensor(1)
    assert model._do_all_gather(torch.tensor(1)) == torch.tensor(1)
    assert model._do_broadcast(torch.tensor(1), src=0) == torch.tensor(1)
    model.barrier()


def test__encode_str__decode_str():
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    s = "test-abcedfg"

    encoded_s = ComputationModel._encode_str(s, device, 1024)
    assert isinstance(encoded_s, torch.Tensor) and encoded_s.shape == (1, 1025)

    decoded_s = ComputationModel._decode_str(encoded_s)
    assert isinstance(decoded_s, list) and len(decoded_s) == 1
    assert decoded_s[0] == s


def test__encode_input_data():
    encoded_msg = ComputationModel._encode_input_data(None, is_src=True)
    assert encoded_msg == [-1] * 512

    encoded_msg = ComputationModel._encode_input_data(12.0, is_src=True)
    assert encoded_msg == [1] + [-1] * 511

    encoded_msg = ComputationModel._encode_input_data("abc", is_src=True)
    assert encoded_msg == [2] + [-1] * 511

    t = torch.rand(2, 512, 32, 32, 64)
    encoded_msg = ComputationModel._encode_input_data(t, is_src=True)
    dtype_str = str(t.dtype)
    true_msg = [0, 5, 2, 512, 32, 32, 64, len(dtype_str), *list(bytearray(dtype_str, "utf-8"))]
    assert encoded_msg == true_msg + [-1] * (512 - len(true_msg))

    t = torch.randint(-1235, 1233, size=(2, 512, 32, 32, 64))
    encoded_msg = ComputationModel._encode_input_data(t, is_src=True)
    dtype_str = str(t.dtype)
    true_msg = [0, 5, 2, 512, 32, 32, 64, len(dtype_str), *list(bytearray(dtype_str, "utf-8"))]
    assert encoded_msg == true_msg + [-1] * (512 - len(true_msg))

    t = torch.tensor(12)
    encoded_msg = ComputationModel._encode_input_data(t, is_src=True)
    dtype_str = str(t.dtype)
    true_msg = [0, 0, len(dtype_str), *list(bytearray(dtype_str, "utf-8"))]
    assert encoded_msg == true_msg + [-1] * (512 - len(true_msg))

    for t in [None, "abc", torch.rand(2, 512, 32, 32, 64), 12.34, object()]:
        encoded_msg = ComputationModel._encode_input_data(t, is_src=False)
        assert encoded_msg == [-1] * 512


def test__decode_as_placeholder():
    device = torch.device("cpu")

    encoded_msg = [-1] * 512
    encoded_msg[0] = 1
    res = ComputationModel._decode_as_placeholder(encoded_msg, device)
    assert isinstance(res, float) and res == 0.0

    encoded_msg = [-1] * 512
    encoded_msg[0] = 2
    res = ComputationModel._decode_as_placeholder(encoded_msg, device)
    assert isinstance(res, str) and res == ""

    encoded_msg = [-1] * 512
    encoded_msg[0] = 0
    encoded_msg[1 : 1 + 7] = [6, 2, 3, 4, 5, 6, 7]
    dtype_str = "torch.int64"
    payload = [len(dtype_str), *list(bytearray(dtype_str, "utf-8"))]
    encoded_msg[1 + 7 : 1 + 7 + len(payload)] = payload
    res = ComputationModel._decode_as_placeholder(encoded_msg, device)
    assert isinstance(res, torch.Tensor) and res.dtype == torch.int64 and res.shape == (2, 3, 4, 5, 6, 7)

    encoded_msg = [-1] * 512
    with pytest.raises(RuntimeError, match="Internal error: unhandled dtype"):
        ComputationModel._decode_as_placeholder(encoded_msg, device)

    t = torch.rand(2, 512, 32, 32, 64)
    encoded_msg = ComputationModel._encode_input_data(t, True)
    res = ComputationModel._decode_as_placeholder(encoded_msg, device)
    assert isinstance(res, torch.Tensor) and res.dtype == t.dtype and res.shape == t.shape

    t = torch.tensor(12)
    encoded_msg = ComputationModel._encode_input_data(t, True)
    res = ComputationModel._decode_as_placeholder(encoded_msg, device)
    assert isinstance(res, torch.Tensor) and res.dtype == t.dtype and res.shape == t.shape


def test__setup_placeholder():
    device = torch.device("cpu")

    from ignite.distributed.utils import _model

    for t in [torch.rand(2, 3, 4), "abc", 123.45]:
        data = _model._setup_placeholder(t, device, True)
        assert isinstance(data, type(t))
        if isinstance(data, torch.Tensor):
            assert (data == t).all()
        else:
            assert data == t



import pytest

from ignite.base import Serializable


def test_state_dict():
    s = Serializable()
    with pytest.raises(NotImplementedError):
        s.state_dict()


def test_load_state_dict():
    s = Serializable()
    s.load_state_dict({})



from argparse import ArgumentParser

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch.optim import SGD
from torch.utils.data import DataLoader
from torchvision.datasets import MNIST
from torchvision.transforms import Compose, Normalize, ToTensor

from ignite.engine import create_supervised_evaluator, create_supervised_trainer, Events
from ignite.metrics import Accuracy, Loss

try:
    import visdom
except ImportError:
    raise ModuleNotFoundError("No visdom package is found. Please install it with command: \n pip install visdom")


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=-1)


def get_data_loaders(train_batch_size, val_batch_size):
    data_transform = Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])

    train_loader = DataLoader(
        MNIST(download=True, root=".", transform=data_transform, train=True), batch_size=train_batch_size, shuffle=True
    )

    val_loader = DataLoader(
        MNIST(download=False, root=".", transform=data_transform, train=False), batch_size=val_batch_size, shuffle=False
    )
    return train_loader, val_loader


def create_plot_window(vis, xlabel, ylabel, title):
    return vis.line(X=np.array([1]), Y=np.array([np.nan]), opts=dict(xlabel=xlabel, ylabel=ylabel, title=title))


def run(train_batch_size, val_batch_size, epochs, lr, momentum, log_interval):
    vis = visdom.Visdom()

    # if not vis.check_connection():
    #     raise RuntimeError("Visdom server not running. Please run python -m visdom.server")

    train_loader, val_loader = get_data_loaders(train_batch_size, val_batch_size)
    model = Net()
    device = "cpu"

    if torch.cuda.is_available():
        device = "cuda"

    model.to(device)  # Move model before creating optimizer
    optimizer = SGD(model.parameters(), lr=lr, momentum=momentum)
    trainer = create_supervised_trainer(model, optimizer, F.nll_loss, device=device)
    evaluator = create_supervised_evaluator(
        model, metrics={"accuracy": Accuracy(), "nll": Loss(F.nll_loss)}, device=device
    )

    train_loss_window = create_plot_window(vis, "#Iterations", "Loss", "Training Loss")
    train_avg_loss_window = create_plot_window(vis, "#Iterations", "Loss", "Training Average Loss")
    train_avg_accuracy_window = create_plot_window(vis, "#Iterations", "Accuracy", "Training Average Accuracy")
    val_avg_loss_window = create_plot_window(vis, "#Epochs", "Loss", "Validation Average Loss")
    val_avg_accuracy_window = create_plot_window(vis, "#Epochs", "Accuracy", "Validation Average Accuracy")

    @trainer.on(Events.ITERATION_COMPLETED(every=log_interval))
    def log_training_loss(engine):
        print(
            f"Epoch[{engine.state.epoch}] Iteration[{engine.state.iteration}/{len(train_loader)}] "
            f"Loss: {engine.state.output:.2f}"
        )
        vis.line(
            X=np.array([engine.state.iteration]),
            Y=np.array([engine.state.output]),
            update="append",
            win=train_loss_window,
        )

    @trainer.on(Events.EPOCH_COMPLETED)
    def log_training_results(engine):
        evaluator.run(train_loader)
        metrics = evaluator.state.metrics
        avg_accuracy = metrics["accuracy"]
        avg_nll = metrics["nll"]
        print(
            f"Training Results - Epoch: {engine.state.epoch} Avg accuracy: {avg_accuracy:.2f} Avg loss: {avg_nll:.2f}"
        )
        vis.line(
            X=np.array([engine.state.epoch]), Y=np.array([avg_accuracy]), win=train_avg_accuracy_window, update="append"
        )
        vis.line(X=np.array([engine.state.epoch]), Y=np.array([avg_nll]), win=train_avg_loss_window, update="append")

    @trainer.on(Events.EPOCH_COMPLETED)
    def log_validation_results(engine):
        evaluator.run(val_loader)
        metrics = evaluator.state.metrics
        avg_accuracy = metrics["accuracy"]
        avg_nll = metrics["nll"]
        print(
            f"Validation Results - Epoch: {engine.state.epoch} Avg accuracy: {avg_accuracy:.2f} Avg loss: {avg_nll:.2f}"
        )
        vis.line(
            X=np.array([engine.state.epoch]), Y=np.array([avg_accuracy]), win=val_avg_accuracy_window, update="append"
        )
        vis.line(X=np.array([engine.state.epoch]), Y=np.array([avg_nll]), win=val_avg_loss_window, update="append")

    # kick everything off
    trainer.run(train_loader, max_epochs=epochs)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--batch_size", type=int, default=64, help="input batch size for training (default: 64)")
    parser.add_argument(
        "--val_batch_size", type=int, default=1000, help="input batch size for validation (default: 1000)"
    )
    parser.add_argument("--epochs", type=int, default=10, help="number of epochs to train (default: 10)")
    parser.add_argument("--lr", type=float, default=0.01, help="learning rate (default: 0.01)")
    parser.add_argument("--momentum", type=float, default=0.5, help="SGD momentum (default: 0.5)")
    parser.add_argument(
        "--log_interval", type=int, default=10, help="how many batches to wait before logging training status"
    )
    parser.add_argument("--log_file", type=str, default=None, help="log file to log output to")

    args = parser.parse_args()

    run(args.batch_size, args.val_batch_size, args.epochs, args.lr, args.momentum, args.log_interval)



import os

import filelock

import nltk
import pytest
import rouge as pyrouge
import torch

import ignite.distributed as idist
from ignite.exceptions import NotComputableError
from ignite.metrics.nlp import Rouge
from ignite.metrics.nlp.rouge import compute_ngram_scores, RougeL, RougeN

from . import CorpusForTest


@pytest.fixture(scope="session", autouse=True)
def download_nltk_punkt(worker_id, tmp_path_factory):
    root_tmp_dir = tmp_path_factory.getbasetemp().parent
    while True:
        try:
            with filelock.FileLock(root_tmp_dir / "nltk_download.lock", timeout=0.2) as fn:
                fn.acquire()
                nltk.download("punkt")
                fn.release()
                break
        except filelock._error.Timeout:
            pass


corpus = CorpusForTest()


@pytest.mark.parametrize(
    "candidate, reference, n, expected_precision, expected_recall",
    [
        ([], [], 1, 0, 0),
        ("abc", "ab", 1, 2 / 3, 2 / 2),
        ("abc", "ab", 2, 1 / 2, 1 / 1),
        ("abc", "ab", 3, 0, 0),
        ("ab", "abc", 1, 2 / 2, 2 / 3),
        ("ab", "cde", 1, 0 / 2, 0 / 3),
        ("aab", "aace", 1, 2 / 3, 2 / 4),
        ("aa", "aaa", 1, 2 / 2, 2 / 3),
        ("aaa", "aa", 1, 2 / 3, 2 / 2),
    ],
)
def test_compute_ngram_scores(candidate, reference, n, expected_precision, expected_recall):
    scores = compute_ngram_scores(candidate, reference, n=n)
    assert pytest.approx(scores.precision()) == expected_precision
    assert pytest.approx(scores.recall()) == expected_recall


def test_wrong_inputs():
    with pytest.raises(ValueError, match=r"ngram order must be greater than zero"):
        RougeN(ngram=0)

    with pytest.raises(ValueError, match=r"alpha must be in interval \[0, 1\]"):
        RougeN(alpha=-1)

    with pytest.raises(ValueError, match=r"alpha must be in interval \[0, 1\]"):
        RougeN(alpha=2)

    with pytest.raises(ValueError, match=r"multiref : valid values are \['best', 'average'\] "):
        RougeN(multiref="")

    with pytest.raises(ValueError, match=r"variant must be 'L' or integer greater to zero"):
        Rouge(variants=["error"])

    with pytest.raises(NotComputableError):
        RougeL().compute()

    with pytest.raises(ValueError):
        Rouge(multiref="unknown")


@pytest.mark.parametrize(
    "ngram, candidate, reference, expected",
    [
        (1, [1, 2, 3], [1, 2], (2 / 3, 2 / 2)),
        (2, [1, 2, 3], [1, 2], (1 / 2, 1 / 1)),
        (1, "abcdef", "zbdfz", (3 / 6, 3 / 5)),
        (2, "abcdef", "zbdfz", (0, 0)),
    ],
)
def test_rouge_n_alpha(ngram, candidate, reference, expected, available_device):
    for alpha in [0, 1, 0.3, 0.5, 0.8]:
        rouge = RougeN(ngram=ngram, alpha=alpha, device=available_device)
        rouge.update(([candidate], [[reference]]))
        results = rouge.compute()
        assert results[f"Rouge-{ngram}-P"] == expected[0]
        assert results[f"Rouge-{ngram}-R"] == expected[1]
        try:
            F = expected[0] * expected[1] / ((1 - alpha) * expected[0] + alpha * expected[1])
        except ZeroDivisionError:
            F = 0
        assert results[f"Rouge-{ngram}-F"] == F


@pytest.mark.parametrize(
    "candidates, references", [corpus.sample_1, corpus.sample_2, corpus.sample_3, corpus.sample_4, corpus.sample_5]
)
def test_rouge_metrics(candidates, references, available_device):
    for multiref in ["average", "best"]:
        # PERL 1.5.5 reference
        apply_avg = multiref == "average"
        apply_best = multiref == "best"
        evaluator = pyrouge.Rouge(
            metrics=["rouge-n", "rouge-l"],
            max_n=4,
            apply_avg=apply_avg,
            apply_best=apply_best,
            alpha=0.5,
            stemming=False,
            ensure_compatibility=False,
        )
        scores = evaluator.get_scores(candidates, references)

        lower_split_references = [
            [ref.lower().split() for ref in refs_per_candidate] for refs_per_candidate in references
        ]

        lower_split_candidates = [candidate.lower().split() for candidate in candidates]

        m = Rouge(variants=[1, 2, 4, "L"], multiref=multiref, alpha=0.5, device=available_device)
        assert m._device == torch.device(available_device)
        m.update((lower_split_candidates, lower_split_references))
        results = m.compute()

        for key in ["1", "2", "4", "L"]:
            assert pytest.approx(results[f"Rouge-{key}-R"], abs=1e-4) == scores[f"rouge-{key.lower()}"]["r"]
            assert pytest.approx(results[f"Rouge-{key}-P"], abs=1e-4) == scores[f"rouge-{key.lower()}"]["p"]
            assert pytest.approx(results[f"Rouge-{key}-F"], abs=1e-4) == scores[f"rouge-{key.lower()}"]["f"]


def _test_distrib_integration(device):
    from ignite.engine import Engine

    rank = idist.get_rank()

    size = len(corpus.chunks)

    data = []
    for c in corpus.chunks:
        data += idist.get_world_size() * [c]

    def update(_, i):
        candidate, references = data[i + size * rank]
        lower_split_references = [reference.lower().split() for reference in references[0]]
        lower_split_candidate = candidate[0].lower().split()
        return [lower_split_candidate], [lower_split_references]

    def _test(metric_device):
        engine = Engine(update)
        m = Rouge(variants=[1, 2, "L"], alpha=0.5, device=metric_device)
        m.attach(engine, "rouge")

        engine.run(data=list(range(size)), max_epochs=1)

        assert "rouge" in engine.state.metrics

        evaluator = pyrouge.Rouge(
            metrics=["rouge-n", "rouge-l"],
            max_n=4,
            apply_avg=True,
            apply_best=False,
            alpha=0.5,
            stemming=False,
            ensure_compatibility=False,
        )
        rouge_1_f, rouge_2_f, rouge_l_f = (0, 0, 0)
        for candidate, references in data:
            scores = evaluator.get_scores(candidate, references)
            rouge_1_f += scores["rouge-1"]["f"]
            rouge_2_f += scores["rouge-2"]["f"]
            rouge_l_f += scores["rouge-l"]["f"]
        assert pytest.approx(engine.state.metrics["Rouge-1-F"], abs=1e-4) == rouge_1_f / len(data)
        assert pytest.approx(engine.state.metrics["Rouge-2-F"], abs=1e-4) == rouge_2_f / len(data)
        assert pytest.approx(engine.state.metrics["Rouge-L-F"], abs=1e-4) == rouge_l_f / len(data)

    _test("cpu")

    if device.type != "xla":
        _test(idist.device())


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



import math
import warnings
from unittest.mock import MagicMock

import pytest
import torch

from ignite.engine import Engine, Events, State

from ignite.handlers.neptune_logger import (
    global_step_from_engine,
    GradsScalarHandler,
    NeptuneLogger,
    NeptuneSaver,
    OptimizerParamsHandler,
    OutputHandler,
    WeightsScalarHandler,
)


def assert_logger_called_once_with(logger, key, value):
    result = logger[key].fetch_values()
    assert len(result.value) == 1

    if isinstance(result.value[0], float):
        assert math.isclose(result.value[0], value, abs_tol=0.01)
    else:
        assert result.value[0] == value


def test_optimizer_params_handler_wrong_setup():
    with pytest.raises(TypeError):
        OptimizerParamsHandler(optimizer=None)

    optimizer = MagicMock(spec=torch.optim.Optimizer)
    handler = OptimizerParamsHandler(optimizer=optimizer)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler OptimizerParamsHandler works only with NeptuneLogger"):
        handler(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_optimizer_params():
    optimizer = torch.optim.SGD([torch.tensor(0.0)], lr=0.01)
    wrapper = OptimizerParamsHandler(optimizer=optimizer, param_name="lr")
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 123

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)
    assert_logger_called_once_with(logger, "lr/group_0", 0.01)
    logger.stop()

    wrapper = OptimizerParamsHandler(optimizer, param_name="lr", tag="generator")
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)
    assert_logger_called_once_with(logger, "generator/lr/group_0", 0.01)
    logger.stop()


def test_output_handler_with_wrong_logger_type():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler OutputHandler works only with NeptuneLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_output_handler_output_transform():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.output = 12345
    mock_engine.state.iteration = 123

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)
    assert_logger_called_once_with(logger, "tag/output", 12345)
    logger.stop()

    wrapper = OutputHandler("another_tag", output_transform=lambda x: {"loss": x})
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)
    assert_logger_called_once_with(logger, "another_tag/loss", 12345)
    logger.stop()


def test_output_handler_metric_names():
    wrapper = OutputHandler("tag", metric_names=["a", "b"])
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 12.23)
    assert_logger_called_once_with(logger, "tag/b", 23.45)
    logger.stop()

    wrapper = OutputHandler("tag", metric_names=["a"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": torch.tensor([0.0, 1.0, 2.0, 3.0])})
    mock_engine.state.iteration = 5

    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    for key, val in [("tag/a/0", 0.0), ("tag/a/1", 1.0), ("tag/a/2", 2.0), ("tag/a/3", 3.0)]:
        assert_logger_called_once_with(logger, key, val)
    logger.stop()

    wrapper = OutputHandler("tag", metric_names=["a", "c"])

    mock_engine = MagicMock()
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine.state = State(metrics={"a": 55.56, "c": "Some text"})
    mock_engine.state.iteration = 7

    with pytest.warns(UserWarning):
        wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 55.56)
    logger.stop()

    # all metrics
    wrapper = OutputHandler("tag", metric_names="all")
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 12.23)
    assert_logger_called_once_with(logger, "tag/b", 23.45)
    logger.stop()

    # log a torch tensor (ndimension = 0)
    wrapper = OutputHandler("tag", metric_names="all")
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": torch.tensor(12.23), "b": torch.tensor(23.45)})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 12.23)
    assert_logger_called_once_with(logger, "tag/b", 23.45)
    logger.stop()

    wrapper = OutputHandler("tag", metric_names="all")
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State(
        metrics={
            "a": 123,
            "b": {"c": [2.34, {"d": 1}]},
            "c": (22, [33, -5.5], {"e": 32.1}),
        }
    )
    mock_engine.state.iteration = 5

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 123),
    assert_logger_called_once_with(logger, "tag/b/c/0", 2.34),
    assert_logger_called_once_with(logger, "tag/b/c/1/d", 1),
    assert_logger_called_once_with(logger, "tag/c/0", 22),
    assert_logger_called_once_with(logger, "tag/c/1/0", 33),
    assert_logger_called_once_with(logger, "tag/c/1/1", -5.5),
    assert_logger_called_once_with(logger, "tag/c/2/e", 32.1),

    logger.stop()


def test_output_handler_both():
    wrapper = OutputHandler("tag", metric_names=["a", "b"], output_transform=lambda x: {"loss": x})
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)

    assert_logger_called_once_with(logger, "tag/a", 12.23)
    assert_logger_called_once_with(logger, "tag/b", 23.45)
    assert_logger_called_once_with(logger, "tag/loss", 12345)
    logger.stop()


def test_output_handler_with_wrong_global_step_transform_output():
    def global_step_transform(*args, **kwargs):
        return "a"

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    with pytest.raises(TypeError, match="global_step must be int"):
        wrapper(mock_engine, logger, Events.EPOCH_STARTED)

    logger.stop()


def test_output_handler_with_global_step_from_engine():
    mock_another_engine = MagicMock()
    mock_another_engine.state = State()
    mock_another_engine.state.epoch = 10
    mock_another_engine.state.output = 12.345

    wrapper = OutputHandler(
        "tag",
        output_transform=lambda x: {"loss": x},
        global_step_transform=global_step_from_engine(mock_another_engine),
    )

    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 1
    mock_engine.state.output = 0.123

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)
    assert_logger_called_once_with(logger, "tag/loss", mock_engine.state.output)

    mock_another_engine.state.epoch = 11
    mock_engine.state.output = 1.123

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)

    result = logger["tag/loss"].fetch_values()
    assert len(result.value) == 2
    assert result.value[1] == mock_engine.state.output

    logger.stop()


def test_output_handler_with_global_step_transform():
    def global_step_transform(*args, **kwargs):
        return 10

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)
    assert_logger_called_once_with(logger, "tag/loss", 12345)

    logger.stop()


def test_output_handler_state_attrs():
    wrapper = OutputHandler("tag", state_attributes=["alpha", "beta", "gamma"])
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 5
    mock_engine.state.alpha = 3.899
    mock_engine.state.beta = torch.tensor(12.23)
    mock_engine.state.gamma = torch.tensor([21.0, 6.0])

    wrapper(mock_engine, logger, Events.ITERATION_STARTED)

    assert_logger_called_once_with(logger, "tag/alpha", 3.899)
    assert_logger_called_once_with(logger, "tag/beta", 12.23)
    assert_logger_called_once_with(logger, "tag/gamma/0", 21.0)
    assert_logger_called_once_with(logger, "tag/gamma/1", 6.0)

    logger.stop()


def test_weights_scalar_handler_wrong_setup():
    with pytest.raises(TypeError, match="Argument model should be of type torch.nn.Module"):
        WeightsScalarHandler(None)

    model = MagicMock(spec=torch.nn.Module)
    with pytest.raises(TypeError, match="Argument reduction should be callable"):
        WeightsScalarHandler(model, reduction=123)

    with pytest.raises(TypeError, match="Output of the reduction function should be a scalar"):
        WeightsScalarHandler(model, reduction=lambda x: x)

    wrapper = WeightsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler WeightsScalarHandler works only with NeptuneLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_weights_scalar_handler(dummy_model_factory):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = WeightsScalarHandler(model, tag=tag)
        logger = NeptuneLogger(
            project="tests/dry-run",
            mode="debug",
        )
        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert_logger_called_once_with(logger, tag_prefix + "weights_norm/fc1/weight", 0.0)
        assert_logger_called_once_with(logger, tag_prefix + "weights_norm/fc1/bias", 0.0)
        assert_logger_called_once_with(logger, tag_prefix + "weights_norm/fc2/weight", 12.0)
        assert_logger_called_once_with(logger, tag_prefix + "weights_norm/fc2/bias", math.sqrt(12.0))

        logger.stop()

    _test()
    _test(tag="tag")


def test_weights_scalar_handler_frozen_layers(dummy_model_factory):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=True)

    wrapper = WeightsScalarHandler(model)
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)

    assert_logger_called_once_with(logger, "weights_norm/fc2/weight", 12.0)
    assert_logger_called_once_with(logger, "weights_norm/fc2/bias", math.sqrt(12.0))

    assert not logger.exists("weights_norm/fc1/weight")
    assert not logger.exists("weights_norm/fc1/bias")

    logger.stop()


def test_grads_scalar_handler_wrong_setup():
    with pytest.raises(TypeError, match="Argument model should be of type torch.nn.Module"):
        GradsScalarHandler(None)

    model = MagicMock(spec=torch.nn.Module)
    with pytest.raises(TypeError, match="Argument reduction should be callable"):
        GradsScalarHandler(model, reduction=123)

    wrapper = GradsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(TypeError, match="Handler GradsScalarHandler works only with NeptuneLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_grads_scalar_handler(dummy_model_factory, norm_mock):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = GradsScalarHandler(model, reduction=norm_mock, tag=tag)
        logger = NeptuneLogger(
            project="tests/dry-run",
            mode="debug",
        )
        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5
        norm_mock.reset_mock()

        wrapper(mock_engine, logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert logger.exists(tag_prefix + "grads_norm/fc1/weight")
        assert logger.exists(tag_prefix + "grads_norm/fc1/bias")
        assert logger.exists(tag_prefix + "grads_norm/fc2/weight")
        assert logger.exists(tag_prefix + "grads_norm/fc2/bias")

        logger.stop()

    _test()
    _test(tag="tag")


def test_grads_scalar_handler_frozen_layers(dummy_model_factory, norm_mock):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=True)

    wrapper = GradsScalarHandler(model, reduction=norm_mock)
    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    norm_mock.reset_mock()

    wrapper(mock_engine, logger, Events.EPOCH_STARTED)

    assert logger.exists("grads_norm/fc2/weight")
    assert logger.exists("grads_norm/fc2/bias")

    assert not logger.exists("grads_norm/fc1/weight")
    assert not logger.exists("grads_norm/fc1/bias")

    logger.stop()


def test_integration():
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    trainer = Engine(update_fn)

    npt_logger = NeptuneLogger(mode="offline")

    def dummy_handler(engine, logger, event_name):
        global_step = engine.state.get_event_attrib_value(event_name)
        logger["test_value"].append(global_step, step=global_step)

    npt_logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

    trainer.run(data, max_epochs=n_epochs)
    npt_logger.close()


def test_integration_as_context_manager():
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    with NeptuneLogger(mode="offline") as npt_logger:
        trainer = Engine(update_fn)

        def dummy_handler(engine, logger, event_name):
            global_step = engine.state.get_event_attrib_value(event_name)
            logger["test_value"].append(global_step, step=global_step)

        npt_logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

        trainer.run(data, max_epochs=n_epochs)


def test_neptune_saver_serializable(dirname):
    mock_logger = MagicMock(spec=NeptuneLogger)
    mock_logger.upload = MagicMock()
    model = torch.nn.Module()
    to_save_serializable = {"model": model}

    saver = NeptuneSaver(mock_logger)
    fname = dirname / "test.pt"
    saver(to_save_serializable, fname)

    assert mock_logger[dirname / "test.pt"].upload.call_count == 1


@pytest.mark.parametrize("model, serializable", [(lambda x: x, False), (torch.nn.Module().to("cpu"), True)])
def test_neptune_saver(model, serializable):
    mock_logger = MagicMock(spec=NeptuneLogger)
    mock_logger.upload = MagicMock()

    to_save_non_serializable = {"model": model}

    saver = NeptuneSaver(mock_logger)
    fname = "test.pt"
    try:
        with warnings.catch_warnings():
            # Ignore torch/serialization.py:292: UserWarning: Couldn't retrieve source code for container of type
            # DummyModel. It won't be checked for correctness upon loading.
            warnings.simplefilter("ignore", category=UserWarning)
            saver(to_save_non_serializable, fname)
    except Exception:
        pass

    assert mock_logger["model"].upload.call_count == int(serializable)


def test_logs_version():
    from ignite import __version__
    from ignite.handlers.neptune_logger import _INTEGRATION_VERSION_KEY

    logger = NeptuneLogger(
        project="tests/dry-run",
        mode="debug",
    )
    assert logger[_INTEGRATION_VERSION_KEY].fetch() == __version__
